// version info:  1.0  -- first alpha release
//                1.0a -- added comments to script files, the increment operator,
//                        and \~0-9 substitutions in s// commands.
//                1.0b -- added y// command
//                1.0c -- added the -M option to go with the M command
//                1.0d -- adds -D options to define the delimiter for -M files
//                1.0e -- adds the command named t which expands tabs
//                1.0f -- adds \n, \r, \t, \e, \b, \a, \s interpretation in
//                        the substitution strings for the s command.
//                        It also adds the x command.
//                1.0g -- adds command S to split strings on a character or
//                        a number of bytes.
//                1.0h -- maintenance release only -- added valgrind support
//                1.0i -- Changed the 'c' command to mean cut, added 'Q' to exit scripts.
//                1.0j -- add escape character support to q, y and other set
//                        commands.
//                1.0k -- added the F command to read files
//                1.0l -- added the W command and the SR command.
//                1.0m -- aded support for "-f -"
//                1.0n -- added support for [digit]cmd -- to allow conditional
//                        execution based on regex patterns stored in registers
//                1.0o -- added Def and C.
//                1.0p -- cleanup help.
//                1.0q -- add the L command to get the file name into a register
//                        and support other characters that / in substitutions
//                1.0r -- help mesage cleanup
//                1.0s -- add a second regex to the pattern range conditional statement
//                1.0t -- change SW01 to SW012
//                1.0v -- if statements
//                1.0w -- extended two pattern conditionals to support numbers and $
//                1.0x -- added the R command to read whole files
//                1.0z -- fix bug in R commandd
//                1.0B -- add prefix to the p command.
//                1.0C -- -l option simplifications
//                1.0D -- documentation changes only
//                1.0F -- -r[0-9] options
//                1.0G -- N command

char const *versionString="1.0G";  // STOP:  keep this in sync with the comment above!

//
// Copyright 2007-2008, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.

#include <iostream>
#include <string>
#include <list>
#include <fstream>
#include <iomanip>
#include <ctype.h>
#include <simple_regex.h>
#include <xlate.h>
#include <map>
#include <cutset.h>
#include <algorithm>

using namespace std;

static string logText(string const &s, size_t max=40);

extern string expand_tabs(char const *s, bool *non_printable_chars, char replacement );

bool logging=false;

char const *programName="[unspecified]";

string stringRegisters[10];

typedef map<string,string> Mapping;

Mapping mapping; // the global textual translation map -- see the -M option.

char  mappingDelimiter='|';  // used to split lines in the map file

static void yelt(char const *filename);  
  //
  // apply yelt commands to the specified file
  // if the filename is 0, read stdin.
  //

static void handleEscapeSequence(string::const_iterator &first,
                                 string::const_iterator &last,
                                 string &substStringm,
                                 bool    leaveSlash = true
                                );
  //
  // handle escape sequences by advancing first past the
  // sequence and appending the correct text to substString.
  // 
  // Note that first should point immediately after the \ and
  // the slash should not have been appended to subststring
  //
  // This function advances first past the escape sequence.
  //
  // If the leaveSlash flag is true, the function will insert a \
  // into the output text for un-interpreted escape sequences --
  // that is, if you do something like this:  \Q
  //
  // With leaveSlash, you'll get \Q as an output.  If leaveSlash is
  // false, you'll just Q.
  // 


static void extendDefaultScript(char const *str);
  //
  //  Append command found in str to the default script
  //  which looks something like this:
  //
  //    w { n; [-e extensions] ; p; }
  //
  //  Where w means, forever repeat the block following, and n means read the next
  //  line of input into string variable 0, and p means print the contents of
  //  string variable 0.  extensions are commands appended to the default script
  //  using the -e option.
  //

static void finalizeDefaultScript();

string script; // default script

static void printHelpAndQuit();

static void readScriptFile(char const *filename);
static void readMapFile(char const *filename);  // read it into "mapping".

static void checkMissingArgument(char const *before, char const *after)
{
  if(*after)
    return; // no problem

  cerr << programName << " Error:  missing argument after " << before << '\n';
  exit(1);
}

int main(int argc, char **argv)
{
  
  programName = *argv++;

  while(*argv)
  {
    if(argv[0][0] == '-')
    {
      char const *arg = *argv++;

      static string dashH             ("-h");
      static string dashdashHelp      ("--help");
      static string dashE             ("-e");
      static string dashSS            ("-S"); // SS => BIG S, not little s
      static string dashF             ("-f"); // specify input script
      static string dashV             ("-v"); // print program version
      static string dashL             ("-l"); // turn on logging (bigtime!)
      static string dashMM            ("-M"); // define map table files
      static string dashDD            ("-D"); // define delimiter for -M files

      if(arg[0] == '-')
      {
        if(arg[1] == 'r' && (*argv != 0) )
        {
          // we may have a -r[0-9] option

          if(isdigit(arg[2]))
          {
            int index = arg[2] - '0';

            if(index >= 0 && index <= 9)
            {
              // ok, we have -r3 or something

              arg = *argv++;

              stringRegisters[index] = arg;

              continue;

            }
          }

        }
      }
      
      if(dashL == arg)
      {
        logging=true;
      }
      else
      if(dashV == arg)
      {
        cerr << programName << "  -- Version " << versionString << endl;
        exit(0);
      }
      else
      if(dashH == arg)
      {
        printHelpAndQuit();
      }
      else
      if(dashdashHelp == arg)
      {
        printHelpAndQuit();
      }
      else                                   
      if(dashE == arg)
      {
        checkMissingArgument(arg, *argv);

        extendDefaultScript(*argv++);
      }
      else                                   
      if(dashSS == arg)    
      {
        if(!script.empty())
        {
          cerr << programName << " Error:  combining -e and -f or -S is illegal" << endl;
          exit(1);
        }
        
        checkMissingArgument(arg, *argv);

        script = *argv++;  // set the entire script in one blow
      }
      else                                   
      if(dashDD == arg)    
      {
        //
        //  set the key/value delimiter for the contents of map files
        //
        checkMissingArgument(arg, *argv);

        mappingDelimiter = argv[0][0];

        argv++;  
      }
      else                                   
      if(dashF == arg)    
      {
        if(!script.empty())
        {
          cerr << programName << " Error:  combining -e and -f or -S is illegal" << endl;
          exit(1);
        }

        checkMissingArgument(arg, *argv);

        char const *scriptName = *argv++;

        readScriptFile(scriptName);


      }
      else                                   
      if(dashMM == arg)    
      {
        checkMissingArgument(arg, *argv);

        char const *mapFileName = *argv++;

        readMapFile(mapFileName);

      }
      else
      {
        --argv;

        cerr << programName << " Error:  invalid command line option: '" << *argv << "'\n";
        exit(1);
      }


    }
    else
      break;
  }

  finalizeDefaultScript(); // if needed

  if(logging)
    cerr << "\n@@@" << logText(script) << endl;


  if(*argv == 0)
  {
    // if the user does not specify any file names, read from stdin

    yelt(0);

  }
  else
  {
    // otherwise, process each names file
  
    while(*argv)
    {
      if(strcmp(*argv, "-") == 0)
      {
        ++argv; // treat the file - as being stdin

        yelt(0);
      }
      else
        yelt(*argv++);
    }
  }

  exit(0);

}

string defaultScriptExtensions;

static void finalizeDefaultScript()
  //
  //  Collect all the script options created using -e and turn them
  //  into a real script -- complete with outer while loop which
  //  reads one line of input into variable 0 and finally prints the result of any
  //  textual transformations on variable 0.
  //
{
  if(script.empty())
  {
    if(!stringRegisters[0].empty())
    {
      cerr << programName << " Error:  -r0 option makes no sense using the default script" << endl;
      exit(1);
    }

    if(defaultScriptExtensions.empty())
      script += "w{ n; p; }";
    else
      script += "w{ n; " + defaultScriptExtensions + " p; }";
  }
}

static void logScriptPosition(ostream &stream, string::const_iterator where)
  //
  //  This function is called when parse errors need to be logged.  It prints 
  //  the text of the script up to the point where the problem was detected.
  //
  //  Note that this code relies on the existence of the file static variable,
  //  script and its input iterator, where, is assumed to point into that string.
  //  If this is not true, this code is going to crash!
  //
{
  size_t offset = where - script.begin();

  if(offset < script.size())
  {
    stream << "    " << script.substr(0, offset) << "<====" << endl;
  }
}

static void extendDefaultScript(char const *extension)
  //
  //  This function is called by the main program to append text to the "default script".
  //
{
  if(script.size())
  {
    cerr << programName << " Error:  -e combined with -f or -S is illegal" << endl;
    exit(1);
  }

  defaultScriptExtensions += extension;
  defaultScriptExtensions += "; ";

}

static void printHelpAndQuit()
{
  cerr << 
       "\n"
       "yelt:  a stream text editor similar to sed but with more internal variables,\n"
       "       user definable functions, escape sequence handling, cut program\n"
       "       and tr program features, and simplified parsing support.\n"
       "\n"
       "For more information, see http://www.bordoon.com/yelt\n"
       "\n"
       "Invocation:\n"
       "   yelt [options] [files]\n"
       "\n"
       "   options:\n"
       "     -f file       use script specified in file\n"
       "     -S string     use script specified in string\n"
       "     -e cmd        add 1 extension command to the default script\n"
       "     -l            turn on diagnostic logging\n"
       "     -M file       define string mapping file\n"
       "     -D char       set the map file delimiter for key/value pairs\n"
       "     -r[0-9] str   pre-propulate register with string\n"
       "\n"
       "Yelt's Default script:\n"
       "  w { n; [ -e extensions ] ; p; }\n"
       "\n"
       "\n"
       "Where this script means:\n"
       "  Repeatedly read the next line, execute any\n"
       "  user supplied commands (ie -e options) and print each line.\n"
       "\n"
       "Writing scripts:\n"
       "  Don't forget to put your own 'w { n; ... p; }' wrapper\n"
       "  in your script file or your -S option string!\n"
       "\n"
       "Comments may be added to script files by adding lines that\n"
       "begin with the # character.  These are the only comments supported.\n"
       "\n"
       "The map file is optional and defines the mapping for the 'M' command.\n"
       "The -M option may be used multiple times and the delimiter option. -D\n"
       "can also be used multiple times:  use -D before -M.\n"
       "\n"
       "If the option, '-f -' is specified, then the script is read\n"
       "from stdin.\n"
       "\n"
       ;

  cerr <<
       "\n"
       "Script language overview\n"
       "\nYelt's command language is very similar in look and feel to that\n"
       "of SED.  Yelt has more variables and different commands, but it is\n"
       "very SED-like.  The commands that yelt supports also have an\n"
       "assembly language look and feel:  the names are terse and registers\n"
       "upon which to operate are part of the command.  Commands come in\n"
       "0-register, 1-register, 2-register, and 3 register variants.\n"
       "Yelt commands should usually be terminated by ;'s.  Here are some\n"
       "quick examples:\n"
       "  a01; p3; A98; x34; q3 stuff; L9; l1; A19;\n"
       "Note:  for 1-register commands, if the register is 0, you can leave it out:\n"
       "the command, 'p;', prints register 0.\n"
       "\n"
       ;

  cerr <<
       "\n"
       "Simple Commands:\n"
       "\nNote that variable numbers listed below are just examples\n"
       "There are 10 variables available for any use.  They are numbered\n"
       "0-9.\n"
       "\n"
       "  n        -- read the next line of input into variable 0.\n"
       "              If end of file, get out of the current loop.\n"
       "  n3       -- read the next line of input into variable 3\n"
       "  N[d]     -- push variable d back into the input stream as if it were a line\n"
       "  R4[d] [f]-- read an entire file into a register.  The register\n"
       "              must be specified immediately after the letter R.\n"
       "              The file's name can be specified in either of two\n"
       "              ways -- either as a second register number or as a\n"
       "              separate token in the command ([f]).  The entire\n"
       "              file will be read into the register variable with \n"
       "              embedded new line characters.  Use SC, SR, or SW\n"
       "              to split it up -- if needed.  If no filename is\n"
       "              specified -- either as a second register number or as\n"
       "              a filename, then stdin will be read.\n"
       "  p        -- print variable 0\n"
       "  p2       -- print variable 2\n"
       "  p 'text' -- print text then variable 0\n"
       "  p9 \"xxx\" -- print xxx then variable 9\n"
       "  a01      -- copy variable 0 into variable 1\n"
       "  A01      -- append variable 0 into variable 1\n"
       "  x49      -- swap variables 4 and 9\n"
       "  q9 X     -- copy string 'X' into variable 9\n"
       "              (note: to inject ; use \\;).  ; is the terminator\n"
       "              leading and trailing blanks are removed so use \\s\n"
       "              to force spaces.\n"
       "  l3       -- put the current line number into variable 3\n"
       "  L9       -- put the current file name in variable 9\n"
       "  j3 W     -- justify variable 3 to the left for a width of W\n"
       "  J1 W     -- justify variable 1 to the right for a width of W\n"
       "  +8       -- increment variable 8\n"
       "  -3       -- decrement variable 3\n"
       "  M9!      -- map variable 9 through the table defined by the -M option\n"
       "              That is, convert variable 9 into its mapped value.\n"
       "              The ! is optional and means the following:  if there is no\n"
       "              mapping found, leave the variable alone.\n"
       "  t4       -- expands tabs in variable 4\n"
       "  SW012 =+@ -- splits variable 0 into three parts, keeping the left side in\n"
       "              variable 0 and moving the right hand side into variable 2.\n"
       "              The character from the delimiter list, '=', '+', or '2' will\n"
       "              be left in register variable 1.\n"
       "              The =+@ characters are a string whose contents defines\n"
       "              delimiters on which the split occurs.  The delimiter\n"
       "              actually found is removed in the process and does not appear\n"
       "              in variable 1.  To use space as a delimiter, use \\s.\n"
       "  SC94 10  -- similar to the SW command but splits based on character\n"
       "              counts.  Variable 9 is reduced in size to the first 10\n"
       "              characters, and the deleted characters are moved into\n"
       "              variable 4.\n"
       "  SR012 /r/-- split the string in variable 0 into 3 pieces:\n"
       "\n"
       "                the part up to but not including the regular expression, r\n"
       "                is left in register 0.\n"
       "\n"
       "                the part that matched the expression is stored in register1\n"
       "\n"
       "                the part after the match is left in register 2\n"
       "\n"
       "  c2 20-39 -- cuts out all the data in variable 2 except for the text\n"
       "              in columns 20 through 39, inclusive.  Column 20 then\n"
       "              becomes column 1.\n"
       "  W75      -- write the contents of variable 7 to the file whose name\n"
       "              is found in variable 5.\n"
       "\n"
       ;

  cerr <<
       "String substitutions:\n"
       "  s/a/b/g   -- substitute in variable 0, all a's with b'\n"
       "  s/c/d/1   -- substitute only the first c with d in variable 0\n"
       "  s3/e/f/g  -- substitute all e's in variable 3 with f's\n"
       "\n"
       "  Note: The first string in a substitution is a grep style\n"
       "        regular expression:  meaning that you must use\n"
       "        \\( ... \\), not ( ... ), for grouping.  You must\n"
       "        use \\+, not just + to force at least one repetition\n"
       "        and \\ escapes almost everything else.\n"
       "\n"
       "        The second string in substitution command is mostly\n"
       "        just plain text, but you can use the following special\n"
       "        sequences:  &      means the entire text pattern matching\n"
       "                           the regular expresion.\n"
       "                    \\0     is the same as &\n"
       "                    \\1-9   refers to the part of the source text\n"
       "                           that matched a \\( \\) in the regular\n"
       "                           expression.  The number is the count of \\('s\n"
       "                           from the beginning of the regex (starting\n"
       "                           with 1.\n"
       "                    \\&     injects & into the output\n"
       "                    \\~0-9  injects the contents of variable 0-9\n"
       "                           into the output.\n"
       "\n"
       "        Other delimiters besides / can be used in the s command, they are:\n"
       "          | : # ?\n"
       "\n"
       ;

  cerr <<
       "        To insert control codes or special characters into the\n"
       "        output using a substitution, use one of the following\n"
       "        escape sequences in the RHS of the substitution:\n"
       "\n"
       "          \\n -- newline\n"
       "          \\t -- tab\n"
       "          \\s -- space\n"
       "          \\b -- backspace\n"
       "          \\a -- bel\n"
       "          \\r -- carriate return\n"
       "          \\e -- escape\n"
       "\n"
       ;

  cerr <<
       "\nConditions and program control:\n"
       "  {cmd ; cmd}       -- execute multiple commands in a block\n"
       "  if /r/ c [else x] -- if regex matches, execute cmd c, else executed command x\n"
       "  w{ cmds }         -- repeatedly execute cmds until end of file\n"
       "                       use 'n' or 'b' or 'd' to getout of the loop.\n"
       "  b                 -- break out of the while loop you are in\n"
       "  d                 -- continue with the loop but skip the rest of\n"
       "                       current iteration thereof\n"
       "  Q                 -- return from the current function or script if not in a function\n"
       "  1,5cmd            -- execute cmd only in lines 1 through 5\n"
       "  40,$ cmd          -- execute cmd on all lines 40 and higher\n"
       "  20,/r/ cmd        -- execute cmd on lines after 20 until the line containing\n"
       "                       /r/ -- inclusive\n"
       "  /stuff/cmd        -- execute cmd only if variable 0 contains stuff\n"
       "  /r/,/s/ cmd       -- execute cmd if curline in range /r/ to /s/ where\n"
       "                       r and s are regular expressions defining a group of lines.\n"
       "                       The line containing /s/ is executed unless the cmd checks\n"
       "                       for that and doesn't execute.  For example:\n"
       "                         /r/, /s/ { /s/! cmd; }\n"
       "                       /r/ and /s/ apply to the same variable (defined in /r/~[digit])\n"
       "  /r/,90 cmd        -- execute cmd on lines starting with the first line to contain\n"
       "                       /r/ until line 90 -- inclusive\n"
       "  /r/,$ cmd         -- execute cmd on all lines after the first containing /r --\n"
       "                       inclusive.\n"
       "  /x/!cmd           -- execute cmd only if variable 0 does no have x in it\n"
       "  /y/~3cmd          -- execute cmd only if varible 3 has a y in it\n"
       "  /z/~4!cmd         -- execute cmd only if variale 4 has no z's in it\n"
       "  F3 {script}       -- Open the file whose name is in variable3 and apply\n"
       "                       the script to it -- you must use the n and p commands\n"
       "                       in the script if you want actually process the text.\n"
       "  |31!cmd           -- execute cmd if the text in variable 1 matches the regex\n"
       "                       stored in variable 3. The ! is optional and if given\n"
       "                       it means that the command will execute if the current\n"
       "                       line does NOT match the regex in var 3.\n"
       "\n"
       << endl;

  cerr <<
       "\nFunctions\n"
       "\n"
       "  C fun 1,3; --  Call function 'fun' passing registers 1 and 3 to that\n"
       "                 function which will use them as its registers 0 and 1\n"
       "                 then return any data to the caller back in the caller's\n"
       "                 registers 1 and 3.\n"
       "  Def fun 2; --  Declare function 'fun' which accepts 2 registers as\n"
       "                 parameters and which will return any values in those same\n"
       "                 registers.  See the function's comments.\n"
       "  Q;         --  Return from the function\n"
       ;

  exit(1);
}

static void compileScript();

static void executeScript(std::istream &);

static string currentFileName;


static void yelt(char const *filename)
  //
  //  Perform all yelt processing on a specified file.  Note that the
  //  yelt registers do NOT get reset between files.
  //
{
  // setup to help the 'L' command

  if(filename)
    currentFileName = filename;
  else
    currentFileName.resize(0);


  if(logging)
  {
    cerr << "@@@yelt<" << currentFileName << '>' << endl;
  }


  compileScript(); // if needed

  if(filename == 0)
    executeScript(cin);
  else
  {
    ifstream s(filename);

    if(!s.is_open())
    {
      cerr << programName << " Error:  cannot open file '" << filename << "'\n";
      exit(1);
    }

    executeScript(s);

  }
}

static int inputLineNumber=0;

struct Statement
  //
  //  A script, once compiled, is a list of Statement objects -- each of which
  //  might have sub-statements.  A statement object is executed in the context
  //  of a given standard input file stream.  Each kind of statement has its own
  //  method of execution (execute_v), and outputting a description of the statement.
  //  The generalized execute() method which is part of Statement itself exists
  //  only so that a standard logging technique can be employed.
  //
{
  struct WhileContext
  {
    // every while-loop and function, including the top level function
    // has a WhileContext object associated with it.  A WhileContext allows statements
    // governed by the context to store variables not understood by the script statements
    // as a whole.  The only user of this feature, at this time, is the pattern range
    // statement:
    //
    //    /expr1/,/expr2/ { cmds; }
    //
    // This statement adds a simple flag variable to its while context to keep track of
    // whether or not the statement is currently active -- ie between expr1 and expr2
    // in the input stream.
    //
    // If a statement needs to have a global variable kept outside its body, such as the
    // "currently active flag" mentioned above, just derive from WhileContext::Data to 
    // to create the data object to be stored then new up an object of that type and
    // insert it into the while context.  This should usually be done in the execute_v()
    // methods like this:
    //
    //  void execute_v(istream &s, WhileContext *wc) const
    //  {
    //     struct ActiveFlag: public WhileContext::Data 
    //     { 
    //        bool isActive_; 
    //        ActiveFlag(bool b): isActive_(b) {}
    //     };
    //
    //     //
    //     //  first, check to see if the flag is already stored
    //     //  in the while loop -- if it isn't, add it with a 
    //     //  default value
    //     //
    //     WhileContext::iterator loc = wc->find(this);
    //
    //     if(loc  == wc->end())
    //     {
    //        (*wc)[this] = ActiveFlag(false);
    //        loc = wc->find(this);
    //     }
    //
    //     // now use the value stored in the while context
    //
    //     ActiveFlag * f = dynamic_cast<ActiveFlag*>(loc->second);
    //
    //     if(f->isActive_)
    //     {
    //       DoSomething();
    //       f->isActive_ = CheckForStillActive();
    //     }
    //     else
    //     {
    //       f->isActive_ = CheckForNewlyActive();
    //       DoSomething();
    //       f->isActive_ = CheckForStillActive();
    //     }
    //    
    //  }

    struct Data
    {
      virtual ~Data(){}
    };


    typedef map<Statement const*, Data*>::iterator iterator;

    iterator begin()           { return data_.begin(); }
    iterator end()             { return data_.end();   }

    iterator find(Statement const*s) { return data_.find(s); }

    void     insert(pair<Statement const*, Data*> const &r) { data_.insert(r); }

    Data *operator[] (Statement const*s) { return data_[s];  }

    ~WhileContext()
    {
      map<Statement const*, Data*>::iterator first = data_.begin(), last = data_.end();

      while(first != last)
      {
        delete first->second;
        ++first;
      }
    }

  private:
    map<Statement const*, Data*> data_;

  };


  int execute(istream &s, WhileContext *wc) const
    //
    //  Call the execute_v() method but perform needed wrapper logic -- especially
    //  the handling of the logging of the statement.
    //
    //  When a statement is executed, it must return an integer error code that
    //  indicates what happened.  See ExecuteRC for an explanations of the return
    //  codes, but the basic idea is that most of the time, statements return "ok"
    //  which is zero. 
    //
  {
    if(logging)
      output(cerr);

    int rc = execute_v(s,wc);

    if(logging && rc)
    {
      cerr << "\n  ---=" << rc << "\n";
    }

    return rc;

  }

  virtual int execute_v(std::istream &s, WhileContext* wc) const=0;
    //
    // execute_v() is where the real statement specific execution occurs.
    // Most of the time, the execute_v() methods do not actually use the
    // stream passed in.  Only a few do:  whileStatements and readStatements.
    // While statements terminate when out of data and read statements read
    // from the stream.
    //
    // execute_v should return one of the following values:

    enum ExecuteRC
    {
      ok         = 0,  // keep going in this block of statement
      returning  = 1,  // exit the current script, block, parent statement, and process no more commands!
      continuing = 2,  // don't exit a loop, but do skip the rest of the block (aka 'continue;' in C++
      breaking   = 3,  // exit the loop you are in but not the function or script (aka 'break;' in C++
    };

  virtual void output(std::ostream &s) const =0;

  friend std::ostream& operator<<(std::ostream &s, Statement const &t)
  {
    t.output(s);

    return s;
  }

};

typedef list<Statement*> Script;
  //
  //  Compiled script is a list of statement*'s.
  //

static Script compiledScript;
  //
  //  compiledScript is the file static variable holding the compiled script,
  //  whose text is found in the file static variable 'script'.
  //

static void executeScript(std::istream &s)
  //
  //  The top level script has a different execution method than the blockStatement
  //  which it resembles.  I could probably rewrite this code to require that the
  //  top level script be a block statement, but I haven't done it.
  //
  //  See the blockStatement's execute method for comments about how script execution
  //  works.
  //
{
  Script::const_iterator first = compiledScript.begin(), last = compiledScript.end();

  inputLineNumber = 0;

  Statement::WhileContext wc;

  while(first != last)
  {
    Statement const * curStatement = *first++;

    int rc = curStatement->execute(s,&wc);

    if(rc)
      break; // some wierd error code

  }

}

//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct whileStatement
: public Statement
  //
  //.  Repeatedly process commands in a statement block until end
  //.  of file or other boolean expression fails (if I ever add them)
  //
{
  Statement *child_;

  int execute_v(std::istream &s, WhileContext* wc) const
  {
    WhileContext lwc;

    for(;;)
    {
      int rc = child_->execute(s,&lwc);  // the child will be a blockStatement

      if(rc == returning)
        return returning;

      if(rc == breaking)
        break;

      // if the return code is contining or ok, just restart the block

    }

    return ok;

  };

  void output(std::ostream &s) const
  {
#if 0
    s << "+++w ";
    s << (*child_);
#endif    
  }

  whileStatement(Statement *child)
  : child_(child)
  {
  }

};

//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct blockStatement
: public Statement
  //
  //.  A blockStatement is a collection of statements in a {} delimited list.  The statements
  //.  are executed in order unless one of the statements returns something other
  //.  than ok.  When that happens, the execute method stops executing the
  //.  block and returns the failing return code.  The While statement checks for two
  //.  failing codes:  breaking, and continuing.  It consumes these two.  Any other non-zero
  //.  return code from the current statement's execute() method will cause the whileStatement
  //.  to terminate.  No functions return any such code at the time of this writing (but this
  //.  statement might be out of date ...)
  //
{
  Script children_;

  void addChild(Statement *child)      { children_.push_back(child); }

  int execute_v(std::istream &s, WhileContext* wc) const
  {
    Script::const_iterator first = children_.begin(), last = children_.end();

    int rc = ok;

    while(first != last)
    {
      Statement const *curStatement = *first++;

      rc = curStatement->execute(s,wc);

      if(rc != ok)
        break;
    }

    return rc;
  };

  void output(std::ostream &s) const
  {
#if 0
    s << "+++{\n";

    Script::const_iterator first = children_.begin(), last = children_.end();

    while(first != last)
    {
      Statement const *curStatement = *first++;

      curStatement->output(s);
    }
    
    s << "+++}\n";
#endif
  }

};

//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

list<string> pushedBackLines;


struct readStatement
: public Statement
  //
  //.  A readStatement, when executed, reads a line from stdin into a
  //.  variable defined in the construction of the readStatement.  If
  //.  read is successful, the execute() function returns ok meaning
  //.  that the script can continue.  If the read fails, the execute()
  //.  method returns 'returning' which is to be interepted by all
  //.  calling loops, blocks, statements, and functions as meaning --
  //.  exit without performing any more command processing.
  //
{
  int registerNumber_;

  int execute_v(std::istream &s, WhileContext* wc) const
  {
    string temp = stringRegisters[registerNumber_];

    if(!pushedBackLines.empty())
    {
      // pushed back lines don't count with respect inputLineNumber which
      // must be maintained as an accurate reflection of the input file

      temp = pushedBackLines.front();

      pushedBackLines.erase( pushedBackLines.begin() );
    }
    else
    {
      if(getline(s, temp) )
         ++inputLineNumber;
      else
        return breaking;  // kill this script
    }

    stringRegisters[registerNumber_] = temp;

    if(logging)
    {
      cerr << "  ---=" << logText(temp) << '\n';
    }

     return ok;  // read may run you out of data, but this by itself
                 // won't stop processing
  };

  void output(std::ostream &s) const
  {
    s << "+++n" << registerNumber_ << '\n';
  }

  readStatement(int regNum=0)
  : registerNumber_(regNum)
  {
  }

};

//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct printStatement
: public Statement
  //
  //.  a printStatement prints the contents of a variable whose
  //.  number is passed into the constructor.
  //
{
  int registerNumber_;

  string prefix_;

  int execute_v(std::istream &s, WhileContext* wc) const
  {
    string text = stringRegisters[registerNumber_];

    cout << prefix_ << text << '\n';

    if(logging)
      cout.flush();

    return ok;
  };

  void output(std::ostream &s) const
  {
    s << "+++p" << registerNumber_ << '\n';


  }

  printStatement(string const &prefix, int regNum)
  : registerNumber_(regNum),
    prefix_(prefix)
  {
  }

};



static Statement *parseStatement(string::const_iterator &first, string::const_iterator &last);


static void compileScript()
  //
  //  This function compiles the global variable, script, into the list of statements,
  //  compiledScript.  It does this by parsing statements out of the string script
  //  and appending each to the compiled script.  This is similar to the parsing of
  //  a blockStatement -- but not exactly identical.
  //
{
  if(!compiledScript.empty())
    return;
  
  string::const_iterator first = script.begin(),
                         last  = script.end();

  int count=0;

  while(first != last)
  {
    Statement *parsedStatement = parseStatement(first, last);

    if(parsedStatement == 0)
      break;

    compiledScript.push_back(parsedStatement);

    ++count;

    if(count >10000)
    {
      cerr << programName << " Error:  script too long" << endl;
      exit(1);
    }
  }
}     

inline void skipblanks(string::const_iterator &first, string::const_iterator &last)
  //
  //  skipblanks() moves iterator first passed any blanks but does not move it passed
  //  the end of the string.
  //
{
  while(first != last)
  {
    char c = *first;

    if(!isspace(c))
      break;

    ++first;

  }
}

inline char gobbleToken(string &word, string::const_iterator &first, string::const_iterator &last)
{
  // parse a standard word, a single character, or >=, ==, etc.
  //
  // return 0 for eof, the first character of the token if one is found

  skipblanks(first, last);

  word.resize(0);

  if(first == last)
    return char(0);

  while(first != last)
  {
    char c = *first;

    if( isspace(c) )
    {
      if(word.empty())
        return 0;
      else
        return word[0];
    }
    else
    if( isalnum(c) )
    {
      if(isdigit(c) && word.empty())
      {
        // if the first character of the token is a digit 
        // then scoop up only the digits and return

        word += c;
        ++first;

        while(   (first != last)
              && isdigit(*first)
             )
        {
          word += *first++;
        }

        return word[0];

      }
      else
      {

        word += c;
        ++first;
      }

    }
    else
    if(    (c == '>')
       ||  (c == '<')
       ||  (c == '=')
       ||  (c == '+')
       ||  (c == '-')
       ||  (c == '*')
       ||  (c == '%')
      )
    {
      // handle multi-character tokens

      if(!word.empty())
      {
        
        return word[0];
      }

      word += c;

      ++first;

      if(first == last)
      {
        return c;
      }
      else
      {
        if(*first == '=')
        {
          word += '=';
          ++first;
        }
 
        return word[0];
      }

    }
    else
    if( (c > ' ') && (c < 0x7f) )
    {
      if(!word.empty())
      {
        return word[0];
      }

      word += c;
      ++first;
      return c;
    }
    else
    {
      cerr << programName << " Error:  invalid character 0x" << hex << unsigned(c) << " in script"
           << endl;

      logScriptPosition(cerr,first);

      exit(1);
    }

  }

  if(word.empty())
  {
    cerr << programName << " Error:  bug in yelt.exe at " __FILE__ << " line " << __LINE__ 
         << endl;

    exit(1);
  }
  return word[0];

}

static Statement *parseWriteStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseWhileStatement(string::const_iterator &first, string::const_iterator &last);
static Statement *parseFileStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseReadAllStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseReadStatement(string const & token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseUnReadStatement(string const & token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseTabsStatement(string const & token, string::const_iterator &first, string::const_iterator &last);
static Statement *parsePrintStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseLineStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseJustifyStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseBlockStatement(string::const_iterator &first, string::const_iterator &last);
static Statement *parseSubstStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseRangeConditionalStatement(string::const_iterator &first, string::const_iterator &last);
static Statement *parsePatternConditionalStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseAssignStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseSwapStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseAppendStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseQuoteStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseAddStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseTranslateStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseMapStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseSplitStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseCutStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseRegPatternStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseFunctionStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseCallStatement(string const &token, string::const_iterator &first, string::const_iterator &last);
static Statement *parseIfStatement(string const &token, string::const_iterator &first, string::const_iterator &last);

//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct endBlockStatement
: public Statement
  //
  //.  an endBlockStatement is never executed -- it is defined merely as a convenience in writing
  //.  the parsing code
  //
{
  int execute_v(std::istream &s, WhileContext* wc) const
  {
    cerr << programName << " Error:  bug in yelt.exe at " __FILE__ << " line " << __LINE__ << endl;
    exit(1);
    return 0;
  };

  void output(std::ostream &s) const
  {
    cerr << programName << " Error:  bug in yelt.exe at " __FILE__ << " line " << __LINE__ << endl;
    exit(1);

  }

};

//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct breakStatement
: public Statement
  //
  //.  a breakStatement's sole purpose is to terminate a loop.  Presumably, you don't
  //.  write a script with an unconditional break -- instead, you make the execution
  //.  of the break statement conditional on something:
  //.
  //.     40b  # break on line 40
  //.
  //.     /stuff/b # break if the line contains "stuff"
  //
{
  int execute_v(std::istream &s, WhileContext* wc) const
  {
    return breaking;
  };

  void output(std::ostream &s) const
  {
    s << "+++b\n";
  }

};
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct continueStatement
: public Statement
  //
  //.  a continueStatement's sole purpose is to restart a loop.  Presumably, you don't
  //.  write a script with an unconditional continue -- instead, you make the execution
  //.  of the break statement conditional on something:
  //.
  //.     40c  # skip the rest of block if you get to line 40 (but only line 40
  //.
  //.     /stuff/b # skip the rest of the block if the current line contains "stuff"
  //
{
  int execute_v(std::istream &s, WhileContext* wc) const
  {
    return continuing;
  };

  void output(std::ostream &s) const
  {
    s << "+++d\n";
  }

};
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct quitStatement
: public Statement
  //
  //.  a quitStatement's sole purpose is to terminate a loop.  Presumably, you don't
  //.  write a script with an unconditional break -- instead, you make the execution
  //.  of the break statement conditional on something:
  //.
  //.     40q  # quit the whole script if ever reach line 40
  //.
  //.     /stuff/q # quit if the line contains "stuff"
  //
{
  int execute_v(std::istream &s, WhileContext* wc) const
  {
    return returning;
  };

  void output(std::ostream &s) const
  {
    s << "+++Q\n";
  }

};


static Statement *parseStatement(string::const_iterator &first, string::const_iterator &last)
  //
  //  parseStatement parses statements out of a string and returns a pointer to the newly
  //  allocated statement objects.  The caller has the responsibility for deleting them --
  //  but right now, the destructors don't work just right -- so don't bother.
  //
{
  string token;

  static string ifkeyword("if");

  string::const_iterator savedFirst = first;  

  char firstWordChar;

  // ignore leading blanks and semicolons at the beginning of statement.

  do
  {
    // read and discard leading ';'

    firstWordChar =  gobbleToken(token, first,last);

    if(firstWordChar == ';')
      savedFirst = first;

  }
  while(firstWordChar == ';');

  

  if(firstWordChar == 0)
    return 0;

  // interpret the first character of every statement as a key that
  // uniquely identifies the statement.


  if(firstWordChar == 'C')
  {
    return parseCallStatement(token, first, last);
  }
  else
  if(firstWordChar == 'D' && (token == "Def"))
  {
    return parseFunctionStatement(token, first, last);
  }
  else
  if(firstWordChar == 'F')
    return parseFileStatement(token, first, last);
  else
  if(firstWordChar == 'w')
    return parseWhileStatement(first, last);
  else
  if(firstWordChar == 'n')
    return parseReadStatement(token, first, last);
  else
  if(firstWordChar == 'N')
    return parseUnReadStatement(token, first, last);
  else
  if(firstWordChar == 'p')
    return parsePrintStatement(token, first, last);
  else
  if(   (firstWordChar == 'l')
     || (firstWordChar == 'L')
    )
    return parseLineStatement(token, first, last);
  else
  if(   (firstWordChar == 'j')
     || (firstWordChar == 'J')
    )
    return parseJustifyStatement(token, first, last);
  else
  if(firstWordChar == '{')
    return parseBlockStatement(first, last);
  else
  if(firstWordChar == '}')
    return new endBlockStatement;
  else
  if(firstWordChar == 's')
    return parseSubstStatement(token, first,last);
  else
  if(firstWordChar == 'b')
    return new breakStatement();
  else
  if(firstWordChar == 'Q')
    return new quitStatement();
  else
  if(firstWordChar == 'd') 
    return new continueStatement;
  else
  if(   (firstWordChar == '$')
     || isdigit(firstWordChar)
    )
  {
    first = savedFirst; // this parser handles the WHOLE statement, not just the parts after the first char
    return parseRangeConditionalStatement(first, last);
  }
  else
  if(firstWordChar == '/')
  {
    return parsePatternConditionalStatement(token, first, last);
  }
  else
  if(firstWordChar == 'a')
  {
    return parseAssignStatement(token, first, last);
  }
  else
  if(firstWordChar == 'c')
  {
    return parseCutStatement(token, first, last);
  }
  else
  if(firstWordChar == 'A')
  {
    return parseAppendStatement(token, first, last);
  }
  else
  if(firstWordChar == 'q')
  {
    return parseQuoteStatement(token, first, last);
  }
  else
  if( (firstWordChar == '+') || (firstWordChar == '-') )
  {
    return parseAddStatement(token, first, last);
  }
  else
  if( firstWordChar == 'y')
  {
    return parseTranslateStatement(token, first, last);
  }
  else
  if( firstWordChar == 'M')
  {
    return parseMapStatement(token, first, last);
  }
  else
  if( firstWordChar == 't')
  {
    return parseTabsStatement(token, first, last);
  }
  else
  if( firstWordChar == 'x')
  {
    return parseSwapStatement(token, first, last);
  }
  else
  if( firstWordChar == 'S')
  {
    return parseSplitStatement(token, first, last);
  }
  else
  if( firstWordChar == 'W')
  {
    return parseWriteStatement(token, first, last);
  }
  else
  if( firstWordChar == '|')
  {
    return parseRegPatternStatement(token, first, last);
  }
  else
  if( (firstWordChar == 'i') && (token == ifkeyword) )
  {
    return parseIfStatement(token, first, last);
  }
  else
  if( firstWordChar == 'R')
  {
    return parseReadAllStatement(token, first, last);
  }


  cerr << programName << " Error:  invalid script command token:  " << token << endl;

  logScriptPosition(cerr,first);

  exit(1);

}


static Statement *parseBlockStatement(string::const_iterator &first, string::const_iterator &last)
{
  // assumes the { token is already gone from the stream
  // and parses statements up to the closing }.  Statements are appended to a blockStatement
  // object which is returned.

  Statement *s=0;

  blockStatement * block = new blockStatement;

  for(;;)
  {
    s = parseStatement(first, last);

    if(dynamic_cast<endBlockStatement*>(s))
    {
      delete s;
      return block;
    }
    else
      block->addChild(s);
  }

  cerr << programName << "  Error:  bug in yelt at " __FILE__ << " line " << __LINE__ << endl;
  logScriptPosition(cerr,first);
  exit(1);

  return 0; //  how did you get here?

}

static Statement *parseWhileStatement(string::const_iterator &first, string::const_iterator &last)
{
  // this code assumes that the 'w' has already been removed from the stream.  It parses up
  // the block statement that must follow the 'w'.

  string token;

  string::const_iterator savedFirst = first;

  char c = gobbleToken(token, first, last);

  if(c != '{')
  {
    // handle boolean expression
    // not yet implemented

    // make sure to gobble the { token in this block
  }

  // assumes token currently contains {

  Statement *block = parseBlockStatement(first, last);

  return new whileStatement(block);

}



static Statement *parseReadStatement(string const & token, string::const_iterator &first, string::const_iterator &last)
{
  // read statements consist of a single token, which is the token that triggered the
  // invocatio of this code -- no use of first or last is necessary

  size_t size = token.size();

  bool syntaxError = false;

  if( (size != 1) && (size != 2))
  {
    syntaxError = true;
  }

  int variable=0;

  if(size > 1)
  {
    if(!isdigit(token[1]))
      syntaxError = true;
    else
      variable = token[1]-'0';
  }


  if(syntaxError)
  {
    cerr << programName << " Error:  n-statement syntax error:  expected one of:\n"
         << "    n\n"
         << "    n[digit]\n"
         << " found\n"
         << "    " << token
         << endl;

    logScriptPosition(cerr,first);

    exit(1);
  }

  return new readStatement(variable);


}


static Statement *parsePrintStatement(string const & token, string::const_iterator &first, string::const_iterator &last)
{
  // print statements consist of a single token, which is the token that triggered the
  // invocatio of this code -- no use of first or last is necessary

  size_t size = token.size();

  bool syntaxError = false;

  if( (size != 1) && (size != 2))
  {
    syntaxError = true;
  }

  int variable=0;

  if(size > 1)
  {
    if(!isdigit(token[1]))
      syntaxError = true;
    else
      variable = token[1]-'0';
  }

  string prefix;

  skipblanks(first, last);

  if(first != last)
  {
    if(*first == '"')
    {
      ++first;

      while(first != last)
      {
        if(*first == '"')
        {
          ++first;
          break;
        }
        else
        if(*first == '\\')
        {
          ++first; 
          handleEscapeSequence(first, last, prefix, false);
        }
        else
          prefix += *first++;
      }
    }
    else
    if(*first == '\'')
    {
      ++first;

      while(first != last)
      {
        if(*first == '\'')
        {
          ++first;
          break;
        }
        else
        if(*first == '\\')
        {
          ++first; 
          handleEscapeSequence(first, last, prefix, false);
        }
        else
          prefix += *first++;
      }
    }
    else
    {
      while(   (first != last) 
            && (*first != ';') 
            && (*first != ' ') 
            && (*first != '{') 
            && (*first != '}')
           )
      {
        if(*first == '\\')
        {
          ++first;

          handleEscapeSequence(first, last, prefix, false);
        }
        else
          prefix += *first++;

      }
    }

  }




  if(syntaxError)
  {
    cerr << programName << " Error:  p-statement syntax error:  expected one of:\n"
         << "    p [string]\n"
         << "    p[digit] [string]\n"
         << " Where string is one of:\n"
         << "     nothing\n"
         << "     'text'\n"
         << "     \"text\"\n"
         << " Found\n"
         << "    " << token
         << endl;

    logScriptPosition(cerr,first);

    exit(1);
  }

  return new printStatement(prefix, variable);

}

inline string::const_iterator findInRange(char c, 
                                          string::const_iterator first, 
                                          string::const_iterator const &last
                                         )
{
  //
  //  search in the range first,last for the character c.
  //
  //  findInRange just skips over text without interpreting it -- except to 
  //  skip over escaped character sequences \x without mistaking the escaped
  //  character as terminating the search.
  //


  while( (first != last) && (c != *first) ) 
  {
    if(*first == '\\')
    {
      ++first;  //skip the back slah

      if(first != last)
        ++first; // then skip the protected character
    }
    else
      ++first; 
  }

  return first;
}
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct substStatement
: public Statement
{
  // a substStatment object is a representation of the
  // the command to substitute one expression for another.
  // The source expression is text matched by a regex.  The target string
  // is plain text with optional "special" character sequences.
  // The special sequences allow parts of the source string to
  // be included in the replacement.  The special sequences
  // are:  &     -- same as \0
  //       \0    -- the entire matching source text string
  //       \1-9  -- see below
  //       \~0-9 -- replace the string with a variable
  //       \&    -- leave a & in the output
  //
  // The \1-\9 sequences refer to parts of the source string
  // that matched the regex.  A regex is allowed to have \( ... \)
  // sub-regexes.  The \1-\9 replacement sequences refer to which one
  // of the \('s is being included in the output text.  Note that if you
  // count \('s from the left in the regex, that is the sub-regex being 
  // substituted.

  string regexString_;
  string substString_;
  char   option_;
  int    variable_;

  SimpleRegex compiledExpression_;

  int execute_v(std::istream &s, WhileContext* wc) const
  {
    string &text = stringRegisters[variable_];

    if(compiledExpression_(text))
    {
      string tmp;

      string replacement = substString_;

      if(replacement.find_first_of("~") < replacement.size())
      {
        // look for \~0 patterns and handle them.

        string newReplacement;

        string::const_iterator first = replacement.begin(), last = replacement.end();

        while(first != last)
        {
          char c = *first++;

          if(c != '\\')
          {
            newReplacement += c;
          }
          else
          {
            if(first == last)
              newReplacement += c;
            else
            {
              if(*first == '~')
              {
                ++first;

                if( (first == last) || !isdigit(*first) )
                  newReplacement += "\\~";
                else
                {
                  int var =  *first++ -'0';

                  newReplacement += stringRegisters[var];
                }
                  

              }
              else
              {
                newReplacement += '\\';      // put the \ in the new replacement string
                newReplacement += *first++;
              }
            }
          }

        }

        replacement = newReplacement;

      }



      if(option_ == '1')
        compiledExpression_.replace(tmp, text, replacement, 1);
      else
        compiledExpression_.replace(tmp, text, replacement);

      if(logging)
      {
        cerr << "   src: " << logText(text) << '\n';
        cerr << "   tgt: " << logText(tmp)  << '\n';
      }

      text = tmp;

    }

    return ok;


  };

  void output(std::ostream &s) const
  {
    s << "+++s" << variable_ <<  "/" << regexString_ << "/" << substString_ << "/" << option_ << '\n';
  }

  substStatement(string const &reg, string const &sub, char option, int variable)
  : regexString_(reg),
    substString_(sub),
    option_(option),
    variable_(variable),
    compiledExpression_(reg)
  {
    if(substString_.find_first_of('&') < substString_.size())
    {
      // replace & with \0 in the subst string

      string tmp;

      string::iterator first = substString_.begin(), last = substString_.end();

      while(first != last)
      {
        char c = *first++;

        if( c == '&' )
        {
          tmp += '\\';
          tmp += '0';
        }
        else
        if(c == '\\')
        {
          tmp += c;
          
          if(first != last)
            tmp += *first++;
        }
        else
          tmp += c;

      }

      substString_ = tmp;


    }

    if(!compiledExpression_.ok())
    {
      cerr << programName << " Error:  s-statement regex part is invalid" << endl;

      cerr << "  regex is '" << reg << "'\n";
      exit(1);
    }
  }
};



static Statement *parseSubstStatement(string const &token, string::const_iterator &first, string::const_iterator &last)
{
  // assumes that the leading s has been removed.
  //
  // This function returns a new substStatment based on incoming text of the following form
  //
  //   s[digit]/regex/replacement/[1g]
  //
  // Note that the regex and replacement are interpreted by the returned substStatement, see its 
  // comments for details

  int variable=0;

  if(token.size() > 1)
  {
    if(!isdigit(token[1]))
    {
      cerr << programName << " Error:  s-statement second character should be a digit or /!" << endl;
      logScriptPosition(cerr,first);
      exit(1);
    }

    variable = token[1] - '0';
  }

  skipblanks(first,last);

  // look for the delimiters of the subst command.  Normally, the / character is the
  // delimiter but sometimes you need a lot of /'s in the source or replacement strings
  // so we need to support other characters as delimiters

  char delim='/';

  if(   (first != last)
     && (   (*first == '/')
         || (*first == '|')
         || (*first == ':')
         || (*first == '?')
         || (*first == '#')
        )
    )
  {
    delim = *first;
  }


  if(  (first == last) || (*first != delim) )
  {
    cerr << programName << " Error:  s-statement must begin with 's[digit]/'" << endl;
    cerr                << "         (note |, :, ?, or # can be used instead of /)" << endl;
    logScriptPosition(cerr,first);
    exit(1);
    
  }

  string::const_iterator firstSlash = findInRange(delim, first+1, last);

  if(firstSlash == last)
  {
    cerr << programName << " Error:  s-statement must have 3 " << delim << " characters in it" << endl;
    logScriptPosition(cerr,first);
    exit(1);
  }

  string::const_iterator secondSlash = findInRange(delim, firstSlash+1, last);

  if(secondSlash == last)
  {
    cerr << programName << " Error:  s-statement must begin with have 3 " << delim << " characters in it" << endl;
    logScriptPosition(cerr,first);
    exit(1);
  }

  // first slash points to the end of the regext
  // second slash points to the end of the substitution

  string::const_iterator option = secondSlash + 1;

  if(option == last)
  {
    cerr << programName << " Error:  s-statement must end with either g or 1" << endl;
    logScriptPosition(cerr,first);
    exit(1);
  }



  if( (*option != 'g') && (*option != '1'))
  {
    cerr << programName << " Error:  s-statement must end with either g or 1" << endl;
    logScriptPosition(cerr,first);
    exit(1);
  }

  if(*first != delim)
  {
    cerr << programName << " bug in ylet.exe at " __FILE__ << " line " << __LINE__ << endl;
    logScriptPosition(cerr,first);
    exit(1);
  }


  string regexString;

  ++first;

  while(first != firstSlash)
  {
    if(*first == '\\')
    {
      ++first;

      handleEscapeSequence(first, last, regexString);  // leave the slash in place

    }
    else
      regexString += *first++;
  }

  ++first;

  string substString;

  while(first != secondSlash)
  {
    char c = *first;

    if(c == '\\')
    {
      // we have an escape character in the output text string

      ++first;

      handleEscapeSequence(first, last, substString); // leave uniniterpret slashes in place

    }
    else
      substString += *first++;
  }

  first = option+1; // skip past the end of this statement -- maybe leaving a trailing ; in the stream

  return new substStatement(regexString, substString, *option, variable);


}
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct rangeConditionalStatement
: public Statement
  //
  //.  a rangeConditionalStatement looks like this:
  //.
  //.     100,130s/a/b/g
  //.
  //.  This means that on lines 100 to 130 (inclusive) subsitute a with b globally.
  //.
  //.  Basically, a range conditional statement is just a wrapper around some other
  //.  statement that handles an "if" clause
  //
{
  int lowerBound_;
  int upperBound_;

  static int maxInt_;

  Statement *child_;

  rangeConditionalStatement(int lb, int ub, Statement *c)
  : lowerBound_(lb),
    upperBound_(ub),
    child_(c)
  {
  }

  int execute_v(std::istream &s, WhileContext* wc) const
  {
     if(   (inputLineNumber >= lowerBound_) 
        && (inputLineNumber <= upperBound_)
       )
     {
       return child_->execute(s,wc);
     }

     return ok;
  }    

  void output(ostream &s) const
  {
    s << "+++" << lowerBound_ << ',';

    if(upperBound_ == maxInt_)
      s << "$ ";
    else
      s << upperBound_ << ' ';

    child_->output(s);
  }


};

int rangeConditionalStatement::maxInt_ = 1000000000;


static Statement *parseRangeConditionalStatement(string::const_iterator &first, string::const_iterator &last)
{
  // this function assumes that we are  dealing with a statement of this form:
  //
  //   lineNumber1[,lineNumber2]statement
  //
  // Where lineNumber1 is any valid integer, and lineNumber2 can be a valid integer or '$'
  //
  // Further, *first is assumed to be pointing to the true beginning of the statement --
  // unlike some other parsers that assume that the first character of the statement has
  // been removed.
  //
  //
  // But if we run into this syntax:
  //
  //  linenumber, /regex statement
  //
  // then we will call this a patternConditionalStatement and let that function handle
  // the parsing thereof.

  string::const_iterator savedFirstEntry = first;
  
  skipblanks(first, last);

  string token;

  while(   (first != last)
        && isdigit(*first)
       )
  {
    token += *first++;
  }

  int lowerBound = atoi(token.c_str());

  if(lowerBound <= 0)
  {

    string tmp(savedFirstEntry, last);

    cerr << programName << " Error: invalid line number on range conditional statement\n"
                        << "        need a number > 0 but received this text:  '" << tmp 
                        << "'\n";
    logScriptPosition(cerr,first);
    exit(1);
  }

  skipblanks(first, last);

  if(first == last)
  {
    cerr << programName << " Error:  expected statement or ',number statement' in range\n"
                        << "conditional statement.  The first token was " << token << ", and the\n"
                        << "rest of the script was empty.\n"
                        ;
    logScriptPosition(cerr,first);
    exit(1);

  }

  string nextToken;

  string::const_iterator savedFirst = first;

  char c = gobbleToken(nextToken, first, last);

  if(   (c != ',') 
     && (c != '{')
     && !isalpha(c)
    )
  {
    cerr << programName << " Error:  expected statement or ',number statement' in range\n"
                        << "conditional statement.  The first token was " << token << ", and the\n"
                        << "next token was " << nextToken << endl
                        ;
    logScriptPosition(cerr,first);
    exit(1);
  }

  int upperBound = rangeConditionalStatement::maxInt_; //1 billion

  if(c== ',')
  {
    c = gobbleToken(nextToken, first, last);

    if(c == '/')
    {
      // if the text after the command is a regular expression, then we are in a mixed
      // mode pattern conditional statement instead of the plain old range conditional --
      // so reset the parsing parameters and let the other statement's parser handle that
      // syntax.

      first = savedFirstEntry;

      string token;

      gobbleToken(token, first, last);

      return parsePatternConditionalStatement(token, first, last);

    }



    if(!isdigit(c) && (c != '$'))
    {

      cerr << programName << " Error:  second line number in range conditional statement,\n"
                          << "next token was " << nextToken << endl
                        ;
      logScriptPosition(cerr,first); exit(1);
    }

    if(c != '$')
    {
      upperBound = atoi(nextToken.c_str());
    }

  }
  else
  {
    // only the leading number is supplied, not a true range
    // so, the range is a 1 line only range

    first = savedFirst;

    upperBound = lowerBound;
  }

  return new rangeConditionalStatement(lowerBound, upperBound, parseStatement(first, last) );

}
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct patternConditionalStatement
: public Statement
  //
  //.  a patternConditionalStatement is statement that only executes
  //.  if a specified variable matches a given regular expression.  For
  //.  example:
  //.
  //.     /stuff/  b
  //.
  //.  There are also other syntaxes:
  //.
  //.     /reg1/,/reg2/ cmd      # executes cmd if between a match of reg1 and reg2 (inclusive)
  //.     30,/reg3/ cmd          # executed cmd if after line 30 and before owr when reg3 found
  //.     /reg4/,$  cmd          # executes cmd if after reg4 found and not end of file 
  //.     /reg5/,40  cmd         # exeuctes cmd if after reg5 but before or on line 40
  //.
  //.  The following statement breaks out of the current block if variable 0 contains
  //.  stuff.  If you want to use a different variable use this syntax:
  //.
  //.    /stuff/~8b
  //.
  //.  This syntax means check for stuff in variable 8 before breaking.
  //.
  //.  The logic is inverted when you use a trailing !:
  //.
  //.    /stuff/~3!b
  //.
  //.  The above ugly syntax means break if variable 3 does not contain stuff,
  //.  and the following means break if variable 0 does not contain stuff.
  //.
  //.    /stuff/!b
  //.
  //
{
  string       expression_;
  string       endExpr_;
  SimpleRegex  compiledExpression_;
  SimpleRegex  compiledEndExpression_;
  bool         invertedLogic_;
  bool         invertedEndLogic_;
  bool         range_;
  int          variable_;
  Statement   *child_;
  int          firstLine_;  // valid in this syntax:  1,/fred/ cmd
  int          lastLine_;   // valid in this syntax:  /fred/,[digits or !]

  patternConditionalStatement(string const &expr,
                              bool          invert,
                              int           variable,
                              Statement    *child,
                              string const &endExpr,
                              bool          invertEnd,
                              int           firstLine,
                              int           lastLine
                             )
  :  expression_(expr),
     compiledExpression_(expr),
     invertedLogic_(invert),
     variable_(variable),
     child_(child),
     endExpr_(endExpr),
     compiledEndExpression_(endExpr),
     range_(!endExpr.empty()),
     invertedEndLogic_(invertEnd),
     firstLine_(firstLine),
     lastLine_(lastLine)
  {
  }

  struct ActiveFlag
  : public WhileContext::Data
  {
    bool isActive_;
    bool isDone_;       // done means don't turn this back on because inputLineNumber is > firstLine_

    ActiveFlag(bool b): isActive_(b), isDone_(false) {}

  };

  ActiveFlag* getActiveFlagPtr(WhileContext* wc) const
  {
    WhileContext::iterator loc = wc->find(this);

    if(loc == wc->end())
    {
      wc->insert( pair<Statement const*, ActiveFlag*>(this, new ActiveFlag(false)) );

      loc = wc->find(this);

    }

    return dynamic_cast<ActiveFlag*>(loc->second);

  }

  int rangeExecute(istream &s, WhileContext *wc, ActiveFlag *f) const
  {
    int rv = child_->execute(s, wc);

    // now check to see if we are supposed to exit the "active" mode
    
    if(lastLine_)
      return rv;  // let the "end of active region" logic be handled at the
                  // beginning of execution of the next use of this statement
                  // if there is a lastLine_ specified


    string text=stringRegisters[variable_];
    
    int match = (0 != compiledEndExpression_(text));
    
    if(invertedEndLogic_)
    {
      match = !match;
    }
    
    if(match)
    {
      f->isActive_ = false;
      f->isDone_   = true;
    
    }

    return rv;

  }


  int execute_v(std::istream &s, WhileContext* wc) const
  {
    if(firstLine_ || lastLine_)
    {
      // this statement is one of the following forms:
      //
      //  30, /fred/
      //  /bill/,40

      ActiveFlag *f = getActiveFlagPtr(wc);

      if(   !f->isDone_ 
         && !f->isActive_
        )
      {
        // check to see if we have entered an active region of the input file

        if(   firstLine_ 
           && (inputLineNumber >= firstLine_) 
           )
        {
          f->isActive_ = true;
        }
      }


      if(f->isActive_)
      {
        // and we already in the active range

        if(   lastLine_ 
           && (lastLine_ < inputLineNumber)
          )
        {
          // we are no longer in range

          f->isActive_ = false;
          f->isDone_   = true;
          return ok;  // don't execute the child because we are out of range

        }

      }
    }



    if(range_ || firstLine_ || lastLine_ )
    {
      // we have the two expression form of the statement -- thus we need to keep
      // track of whether or not we are inside the range of lines defined by the
      // two expressions

      ActiveFlag *f = getActiveFlagPtr(wc);

      int rv = ok;

      if(   !f->isActive_ 
         && !firstLine_
        )
      {
        // we are not currently active, and there is no first line spec, and
        // we are not passed the end of the active region

        // check to see if this line turns the active state on

        string text=stringRegisters[variable_];
  
        int match = (0 != compiledExpression_(text));

        if(invertedLogic_)
        {
          match = !match;
        }

        if(match)
        {
          f->isActive_ = true;
        }
        else
        {
        }
      }


      if(f->isActive_)
      {
        return rangeExecute(s, wc, f);
      }

      return rv;


    }
    else
    {
      // this statement has only one condition specified
      // so it doesn't need to worry about ranges -- if the current line
      // passes the test, just execute the child statement
      
      string text=stringRegisters[variable_];
  
      int match = (0 != compiledExpression_(text));
  
      if(invertedLogic_)
      {
        match = !match;
      }
  
  
      if(match)
      {
        return child_->execute(s,wc);
      }
    }

    return ok;
  }

  void output(ostream &s) const
  {
    s << "+++/" << expression_ << "/~" << variable_;


    if(invertedLogic_)
      s << "!";

    s << '\n';

    child_->output(s);
  }



};



static Statement *parsePatternConditionalStatement(string const &token, string::const_iterator &first, string::const_iterator &last)
{
  //
  // This function parses 4 syntaxes:
  //
  //   /expr/           cmd
  //   /expr1/,/expr2/  cmd
  //   [digits],/expr/  cmd 
  //   /expr/,[digits]  cmd
  //
  // Note that the /expr/ can be followed by ~[digit][!] in all cases
  //  
  //
  //
  // A pattern conditional statement looks like this:
  //
  //   /regex/[~digit][!]regularStatement
  //
  // See the comment in the rangeConditionalStatement, above.

  string::const_iterator savedFirst = first;

  if(token.empty())
  {
    cerr << programName << ": Bug at " << __FILE__ << ":" << __LINE__ << "\n";
    logScriptPosition(cerr,first); 
    exit(1);
  }

  int    firstLine=0;
  int    lastLine =0;
  string expression;
  bool   invertedLogic = false;
  int    variable=0;
  


  if(isdigit(token[0]))
  {
    firstLine = atoi(token.c_str());
  }
  else
  {
  
    string::const_iterator slash = findInRange('/', first, last);
  
  
    if(slash == last)
    {
      cerr << programName << " Error:  expected trailing / in pattern conditional statement" << endl;
      logScriptPosition(cerr,first); exit(1);
    }
  
  
    while(first != slash)
    {
      expression += *first++;
    }
  
    ++first; // first is now 1 past slash
  
    if(first == last)
    {
      cerr << programName << " Error:  expected statement after /expr/" << endl;
      logScriptPosition(cerr,first); exit(1);
      
    }
  
    if(*first == '~')
    {
      ++first;
  
      if(first == last)
      {
        cerr << programName << " Error:  expected statement after /expr/" << endl;
        logScriptPosition(cerr,first); exit(1);
      }
  
      if(!isdigit(*first))
      {
        cerr << programName << " Error:  expected variable number after /expr/~" << endl;
        logScriptPosition(cerr,first); exit(1);
      }
  
      variable = *first - '0';
  
      ++first;
      
    }
    
  
  
    if(*first == '!')
    {
      invertedLogic = true;
      ++first;
  
    }
  
    if(first == last)
    {
      cerr << programName << " Error:  expected statement after /expr/" << endl;
      logScriptPosition(cerr,first); exit(1);
      
    }
  }

  // now, lets check to see if we have a second regex which will change this
  // into a range "pattern Range Conditional statement".

  string endRangeRegex;
  bool   invertEndLogic=false;

  skipblanks(first,last);

  if( (first != last) && (*first == ',') )
  {
    // ok, we have something like:
    //
    //   /stuff/~0!,/more/
    //             ^

    ++first;

    skipblanks(first, last);

    if(    (first != last) 
       &&  (*first == '$')
      )
    {
      lastLine = (1<<31) -1;
      ++first;
    }
    else
    if(   (first != last)
       && isdigit(*first)
      )
    {
      // ok, we have a /asdf/,number form

      string numberToken;

      gobbleToken(numberToken, first, last);

      lastLine = atoi(numberToken.c_str());

    }
    else
    if( (first != last) && (*first == '/') )
    {
      ++first;

      string::const_iterator begin = first;

      string::const_iterator slash = findInRange('/', first, last);

      if(slash == last)
      {
        cerr << programName << " Error:  expected trailing / in pattern conditional statement" << endl;
        logScriptPosition(cerr,first); exit(1);
      }

      endRangeRegex.assign(begin,slash);

      first=slash;
      ++first;

      if(first != last)
      {
        if(*first == '~')
        {
          cerr << programName << " Error:  variable spec on end conditional not allowed" << endl;
          logScriptPosition(cerr,first); exit(1);
        }

        if(*first == '!')
        {
          invertEndLogic = true;
          ++first;
        }
      }
    }
  }


  return new patternConditionalStatement(expression, 
                                         invertedLogic, 
                                         variable, 
                                         parseStatement(first, last),
                                         endRangeRegex,
                                         invertEndLogic,
                                         firstLine,
                                         lastLine
                                        ); 

}
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct assignStatement
: public Statement
  //
  //.  An assign statement exists to copy one variable to another.  It has not data
  //.  in the statement other than the first token that looks like this:
  //.
  //.     a49
  //.
  //.  Which means to copy register 4 into register 9.
  //
{
  int from_;
  int to_;

  assignStatement(int from, int to)
  : from_(from),
    to_(to)
  {
  }

  int execute_v(istream &s, WhileContext* wc) const
  {
    stringRegisters[to_] = stringRegisters[from_];
    return ok;
  }

  void output(ostream &s) const
  {
    s << "+++a" << from_ << to_ << '\n';
  }

};


static Statement *parseAssignStatement(string const &token, string::const_iterator &first, string::const_iterator &last)
{
  // an assign statement just copies one variable to another:
  //
  //  a01 copies 0 to 1
  //  a10 copies 1 to 0
  //
  // there are 10 variables, 0-9
  //
  // There are no optional parameters or other parameters outside the single token containing a01

  if(token.size() != 3)
  {
    cerr << programName << " Error:  a-statements must look like this:  a[DIGIT][DIGIT]\n";
    logScriptPosition(cerr,first); exit(1);
  }

  char from = token[1];
  char to   = token[2];

  if(   (from == to)
     || !isdigit(from)
     || !isdigit(to)
    )
  {
    cerr << programName << " Error: a-statements require different variable NUMBERS"; 
    logScriptPosition(cerr,first); exit(1);
  }

  return new assignStatement(from-'0', to-'0');

}
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct quoteStatement
: public Statement
  //
  //  A quote statement just moves constant text to a specific register.
  //
{
  string text_;
  int    variable_;

  quoteStatement(string const &s, int variable): text_(s), variable_(variable) {}

  int execute_v(istream &s, WhileContext* wc) const
  {
    stringRegisters[variable_] = text_;

    return ok;
  }

  void output(ostream &s) const
  {
    s << "+++q " << text_ << '\n';
  }

};



static Statement *parseQuoteStatement(string const &token, string::const_iterator &first, string::const_iterator &last)
{
  int variable=0;

  if(   (token.size() != 2)
     || !isdigit(token[1])
    )
  {
    cerr << programName << " Error:  q-statement must have a variable NUMBER following the q" << endl;
    logScriptPosition(cerr, first);
    exit(1);
  }

  variable = token[1] - '0';


  string::const_iterator semicolon = findInRange(';', first, last);

  if(semicolon == last)
  {
    cerr << programName << " Error:  q-statement must have a ; to terminate it\n";
    logScriptPosition(cerr, semicolon);
    exit(1);
  }

  // remove leading blanks from the text

  skipblanks(first, last);

  // remove trailing blanks

  while(first != semicolon)
  {
    --semicolon;

    if(*semicolon != ' ')
    {
      ++semicolon;
      break;
    }
  }

  string value;

  while(first != semicolon)
  {
    if(*first== '\\')
    {
      ++first;

      handleEscapeSequence(first, last, value, false); 

    }
    else
      value += *first++;
  }

  return new quoteStatement(value, variable);

}



//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct appendStatement
: public Statement
  //
  //.  An append statement exists to append one variable to another:
  //.
  //.    A13
  //.
  //.  This command appends register 1 to the end of register 3.
  //
{
  int from_;
  int to_;

  appendStatement(int from, int to)
  : from_(from),
    to_(to)
  {
  }

  int execute_v(istream &s, WhileContext* wc) const
  {
    stringRegisters[to_] += stringRegisters[from_];
    return ok;
  }

  void output(ostream &s) const
  {
    s << "+++A" << from_ << to_ << '\n';
  }

};


static Statement *parseAppendStatement(string const &token, string::const_iterator &first, string::const_iterator &last)
{
  // an append statement just appends one variable to another:
  //
  //  A01 appends 0 to the end of 1
  //  A10 appends 1 to the end of 0
  //
  // there are 10 variables, 0-9

  if(token.size() != 3)
  {
    cerr << programName << " Error:  A-statements must look like this:  a[DIGIT][DIGIT]\n";
    logScriptPosition(cerr,first); exit(1);
  }

  char from = token[1];
  char to   = token[2];

  if(   (from == to)
     || !isdigit(from)
     || !isdigit(to)
    )
  {
    cerr << programName << " Error: a-statements require different variable NUMBERS"; 
    logScriptPosition(cerr,first); exit(1);
  }

  return new appendStatement(from-'0', to-'0');

}
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct lineStatement
: public Statement
  //
  //.  a lineStatement populates a specified register variable with a string
  //.  representing the current line number.  It is formatted as a 5 character
  //.  field, including a trailing blank.
  //
{
  int registerNumber_;

  char cmdType_; // 'l' or 'L'

  int execute_v(std::istream &s, WhileContext* wc) const
  {
    if(cmdType_ == 'L')
    {
      stringRegisters[registerNumber_] = currentFileName; // see function yelt() above
    }
    else
    {

      char buffer[256];
  
      sprintf(buffer, "%4d ", inputLineNumber);
  
      stringRegisters[registerNumber_] = buffer;
    }
    return ok;
  };

  void output(std::ostream &s) const
  {
    s << "+++" << cmdType_;

    if(registerNumber_)
      s << registerNumber_;

    s << '\n';

  }

  lineStatement(int regNum, char c)
  : registerNumber_(regNum),
    cmdType_(c)
  {
  }

};

static Statement *parseLineStatement(string const & token, string::const_iterator &first, string::const_iterator &last)
{
  // line statements consist of a single token, which is the token that triggered the
  // invocation of this code -- no use of first or last is necessary.  A line statement looks
  // like this:
  //
  //   l3
  //
  // Which puts the current line number in register 3 as a string with a trailing blank.
  //
  // If the command begins with capital L instead of little l, use the current file name
  // istead of the line number
  // 

  size_t size = token.size();

  bool syntaxError = false;

  if( (size != 1) && (size != 2))
  {
    syntaxError = true;
  }

  int variable=0;

  if(size > 1)
  {
    if(!isdigit(token[1]))
      syntaxError = true;
    else
      variable = token[1]-'0';
  }


  if(syntaxError)
  {
    cerr << programName << " Error:  l-statement syntax error:  expected one of:\n"
         << "    n\n"
         << "    n[digit]\n"
         << " found\n"
         << "    " << token
         << endl;

    logScriptPosition(cerr,first);

    exit(1);
  }

  return new lineStatement(variable, token[0]);

}


//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct justifyStatement
: public Statement
  //
  //.  a justifyStatement widens a variable either
  //.  to the left or to the right.  Justification is
  //.  controlled by the alphabetic case of the 'j':
  //.
  //.   j1 5  left justifies register 1 in a 5 character field
  //.   j2 10 right justifies register 2 in a 10 character field
  //
{
  int registerNumber_;
  int dir_;
  int width_;

  int execute_v(std::istream &s, WhileContext* wc) const
  {
    string *p = &stringRegisters[registerNumber_];

    while(p->size() < width_)
    {
      if(dir_)
      {
        (*p) = ' ' + (*p);
      }
      else
      {
        (*p) += ' ';
      }
    }
    
    return ok;
  };

  void output(std::ostream &s) const
  {
    if(dir_)
      s << "+++J";
    else
      s << "+++j";

    if(registerNumber_)
      s << registerNumber_;


    s << width_ << '\n';

  }

  justifyStatement(int regNum, int width, int dir)
  : registerNumber_(regNum),
    width_(width),
    dir_(dir)
  {
  }

};

static Statement *parseJustifyStatement(string const & token, string::const_iterator &first, string::const_iterator &last)
{
  //
  //  justify statements look like this:
  //
  //    j9 20
  //    J4 10
  //
  //  The first form left justifies variable 9 in a 20 character field.  The second
  //  right justifies variable 4 in a 10 character field.
  //

  size_t size = token.size();

  bool syntaxError = false;

  int variable=0;
  int dir=0;

  if(token[0] == 'J')
    dir=1;  // right justify

  if(size > 1)
  {
    if(!isdigit(token[1]))
      syntaxError = true;
    else
      variable = token[1]-'0';
  }

  string width;

  char c = gobbleToken(width, first, last);

  if(!isdigit(c))
    syntaxError=true;

  if(syntaxError)
  {
    cerr << programName << " Error:  l-statement syntax error:  expected one of:\n"
         << "    j width\n"
         << "    j[digit] width\n"
         << "    J width\n"
         << "    J[digit] width\n"
         << " found\n"
         << "    " << token << " " << width
         << endl;

    logScriptPosition(cerr,first);

    exit(1);
  }

  return new justifyStatement(variable, atoi(width.c_str()), dir);

}

static void readScriptFile(char const *scriptName)
{
  //
  //  Read the specified script file and append its contents to the global
  //  variable script.  This function reads the file one line at a time and
  //  removes comments.  If the name of the script file is -, then use stdin
  //  instead of opening the script file.
  //

  static string dash("-");

  ifstream *scriptFilePtr=0;

  istream *scriptFile=0;

  if(dash == scriptName)
  {
    scriptFile = &cin;
  }
  else
  {
    scriptFilePtr = new ifstream(scriptName);

    scriptFile = scriptFilePtr;

    if(scriptFilePtr && !scriptFilePtr->is_open())
    {
      // we couldn't open the file

      scriptFile = 0; 

      delete scriptFilePtr;

      // the error message is handled below



    }

  }



  if(scriptFile == 0)
  {
    cerr << programName << " Error:  script file '" << scriptName << "' is invalid\n";
    exit(1);
  }

  script.resize(0);

  while(*scriptFile)
  {
    
    string tmp;

    getline(*scriptFile, tmp);

    string::const_iterator first = tmp.begin(), last = tmp.end();

    string token;

    char c = gobbleToken(token, first, last);

    if(c != '#')
      script += tmp + "\n";
    
  }

  delete scriptFilePtr;


}

//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct addStatement
: public Statement
  //
  //.  and addStatement performs the incrementing or decrementing of
  //.  of a variable by 1.  Statements look like this:
  //.
  //.     +3         -- increment variable 3 by 1
  //.     -          -- decrement variable 0 by 1
  //.
  //.  The produces a fatal error if the variable is not a properly formatted
  //.  integer.
  //
{
  char direction_;  // + or -
  int  variable_ ;


  addStatement(char c, int v)
  : direction_(c),
    variable_(v)
  {
  }

  int execute_v(istream &s, WhileContext* wc) const
  { 
    string text = stringRegisters[variable_];

    if(   text.empty() 
       || !(
               isdigit(text[0]) 
            || text[0] == '-'
           )
      )
    {
      cerr << programName << " Error:  " << direction_ << " statement bad variable reference" << endl;
      cerr <<                "         Variable " << variable_ << " is empty or not a number\n"
           <<                "         it contains '" << text << "'\n";
      exit(1);
    }

    int currentValue = atoi(text.c_str());

    if(direction_ == '+')
      ++currentValue;
    else
      --currentValue;


    char buffer[256];

    sprintf(buffer, "%d", currentValue);

    stringRegisters[variable_] = buffer;

    return ok;
  }

  void output(ostream &s) const
  {
    s << "+++" << direction_ << variable_ << '\n';
  }


};

static Statement *parseAddStatement(string const &token, string::const_iterator &first, string::const_iterator &last)
{
  //
  //  add statements look like this:
  //
  //   +3
  //   -2
  //
  //  These statements increment or decrement the numbered variable by 1.
  //


  if(   (token.size() !=1) 
     || (  
            (token[0] != '+')
         && (token[0] != '-' ) 
        )
    )
  {
    cerr << programName <<  " Error: bug in yelt.exe at " __FILE__ << ':' << __LINE__ << endl;
    
    logScriptPosition(cerr,first);

    exit(1);
    
  }

 int variable=0;

 skipblanks(first, last); // allow spaces after the +/-

 if(first != last)
   variable = *first++ - '0';


 return new addStatement(token[0], variable);


}
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct translateStatement
: public Statement
  //
  //  translate statements look like this:
  //
  //    y[register]/incomingPatternList/outgoingPatternList/
  //
  // Text in the specified register variable (or variable 0 if none is
  // specified) will be converted from what it currently is to what it
  // needs to be based the incomingPatternList and the outgoingPatternList.
  // This works similar to the unix tr command.
  //
{
  string lhs_, rhs_;
  int    variable_;

  TranslateSet translator_;

  translateStatement(string const &lhs, string const &rhs, int variable)
  : lhs_(lhs),
    rhs_(rhs),
    variable_(variable),
    translator_(lhs, rhs)
  {
  }

  int execute_v(istream &s, WhileContext* wc) const
  {
    string text = stringRegisters[variable_];

    string rv = translator_.translate(text);

    stringRegisters[variable_] = rv;

    return ok;
  }


  void output(ostream &s) const
  {
    s << "+++y/" << lhs_ << "/" << rhs_ << "/" << endl;
  }


};

static Statement *parseTranslateStatement(string const &token, string::const_iterator &first, string::const_iterator &last)
{
  //
  //  translate statements look like this:
  //
  //    y[register]/incomingPatternList/outgoingPatternList/
  //
  // Text in the specified register variable (or variable 0 if none is
  // specified) will be converted from what it currently is to what it
  // needs to be based the incomingPatternList and the outgoingPatternList.
  // This works similar to the unix tr command.
  //

  skipblanks(first,last);

  if( (first == last) || (*first != '/') )
  {
     cerr << programName << " Error:  y-statement error, expected / to start the first charset"
     << endl;
     logScriptPosition(cerr, first);
     exit(1);
  }

  if(  (     (token.size() != 1)
         &&  (token.size() != 2)
       )
     ||  (
              (token.size() == 2)
          &&  !isdigit(token[1])
         )
    )
  {
     cerr << programName << " Error:  y-statement error, invalid character after the 'y', expected a digit."
     << endl;
     logScriptPosition(cerr, first);
     exit(1);

  }

 int variable=0;

 if(token.size() > 1)
   variable = token[1] - '0';

 // we know that first points a / at this point.

 string lhs, rhs;

 ++first;

 while(   (first != last)
       && (*first != '/') 
      )
 {
   if(*first == '\\')
   {
     ++first;

     handleEscapeSequence(first, last, lhs, false); // don't leave the slash in the stream

   }
   else
     lhs += *first++;
 }

 if(first == last)
 {
     cerr << programName << " Error:  y-statement error, expected 3 / characters in the command"
     << endl;
     logScriptPosition(cerr, first);
     exit(1);
 }

 ++first;

 while(   (first != last)
       && (*first != '/') 
      )
 {
   if(*first == '\\')
   {
     ++first;

     handleEscapeSequence(first, last, rhs, false); // don't leave the slash in the string

   }
   else
     rhs += *first++;
 }

 if(first == last)
 {
     cerr << programName << " Error:  y-statement error, expected 3 / characters in the command"
     << endl;
     logScriptPosition(cerr, first);
     exit(1);
 }

 ++first;

 return new translateStatement(lhs, rhs, variable);

}

//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct mapStatement
: public Statement
  //
  //  The map statement looks like this:
  //
  //    M1
  //    M2!
  //
  //  This statement has the following meaning:  convert the contents of string
  //  register 1 to the value which is obtained by using those contents as a key
  //  into the global map table, mapping.  This table is populated at the beginning
  //  of yelt execution from the file specified in the -M option.  If no such file is
  //  specified then the mapping will result in an empty string.  If the optional !
  //  is provided, empty strings will not be written back to the variable being mapped
  //  but rather, they will be left unmolested.
  //
{
  int variable_;
  bool keepnulls_;   // if the mapped value is not found, leave the original text alone

    mapStatement(int variable, bool keepnulls)
    : variable_(variable),
      keepnulls_(keepnulls)
    {
    }

    int execute_v(istream &s, WhileContext* wc) const
    {
      string text = stringRegisters[variable_];

      string rv;

      Mapping::const_iterator location = mapping.find(text);

      if(location != mapping.end())
        rv = location->second;
      else
      if(keepnulls_)
        rv = text;

      stringRegisters[variable_] = rv;

      return ok;

    }

    void output(ostream &s) const
    {
      s << "+++M" << variable_ << '\n';
    }
};

static void readMapFile(char const *filename)
{
  //
  //  This function is invoked by the -M option.  It reads the specified file
  //  and splits the lines into two parts based on the mappingDelimiter which
  //  normally defaults to '|' but which can be specified using the -D option
  //  on yelt's command line.
  //
  //  Multiple -M options can be used and the -D option can be changed at any
  //  time.
  //


  ifstream mapFile(filename);

  if(!mapFile.is_open())
  {
    cerr << programName << " Error:  cannot open -M file, " << filename << endl;
    exit(1);
  }

  string line;

  int lineNo=0;

  while(mapFile)
  {
    if(getline(mapFile, line))
    {
      ++lineNo;

      size_t pos = line.find_first_of(mappingDelimiter);

      if(pos >= line.size())
      {
        cerr << programName << " Error:  -M file, " << filename << ", missing delimiter | at line"
             << lineNo
             << endl;

        exit(1);
      }

      string key, value;

      key = line.substr(0,pos);

      value = line.substr(pos+1, line.size() - pos - 1);

      mapping[key] = value;  // first one wins

    }
    else
      break;
  }


}

static Statement *parseMapStatement(string const &token,
                                    string::const_iterator &first,
                                    string::const_iterator &last
                                   )
{
  //
  //  The map statement looks like this:
  //
  //    M1
  //    M2!
  //
  //  This statement has the following meaning:  convert the contents of string
  //  register 1 to the value which is obtained by using those contents as a key
  //  into the global map table, mapping.  This table is populated at the beginning
  //  of yelt execution from the file specified in the -M option.  If no such file is
  //  specified then the mapping will result in an empty string.  If the optional !
  //  is provided, empty strings will not be written back to the variable being mapped
  //  but rather, they will be left unmolested.
  //

  if(   (token.size() > 2)
     || (
            (token.size() == 1)
         && !isdigit(token[1])
        )
    )
  {

    cerr << programName << " Error:  M-statement too long or does not have a digit in column2"<< endl;
    logScriptPosition(cerr, first);

    exit(1);
  }

  int variable = 0;
  bool keep = false;  // if keep is true, don't empty out variables when there is no mapping.

  if(token.size() > 1)
    variable = token[1] - '0';

  skipblanks(first, last);

  if(first != last)
  {
    if(*first == '!')
    {
     ++first;

     keep = true;
    }
  }

  return new mapStatement(variable, keep);



}

//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct tabsStatement
: public Statement
  //
  //.  A tabsStatement, when executed, expands tab characters into the
  //.  appropriate number of spaces in the specified register.
  //
{
  int registerNumber_;

  int execute_v(std::istream &s, WhileContext* wc) const
  {
    string temp = stringRegisters[registerNumber_];

    bool badCharsFound=false;

    string expandedForm = expand_tabs(temp.c_str(), &badCharsFound, ' ');

    stringRegisters[registerNumber_] = expandedForm;

    return ok;  
                 
  };

  void output(std::ostream &s) const
  {
    s << "+++t";

    if(registerNumber_)
      s << registerNumber_;

    s << '\n';
    
  }

  tabsStatement(int regNum=0)
  : registerNumber_(regNum)
  {
  }

};

static Statement *parseTabsStatement(string const &token,
                                     string::const_iterator &first,
                                     string::const_iterator &last
                                    )
{
  int variable=0;


  if(token.size() > 1 )
  {
    if( (token.size() > 2) || !isdigit(token[1]) )
    {
      
      cerr << programName << " Error:  t-statement error, expected t, or t[digit], not "
           << token
           << endl;
      logScriptPosition(cerr, first);
      exit(1);

    }
  }

 return new tabsStatement(variable);

}
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct swapStatement
: public Statement
{
  int lhs_, rhs_;

  swapStatement( int l, int r )
  : lhs_(l),
    rhs_(r)
  {
  }

  int execute_v(istream &s, WhileContext* wc) const
  {
    stringRegisters[lhs_].swap(stringRegisters[rhs_]);

    return ok;
  }

  void output(ostream &s) const
  {
    s << "+++x" << lhs_ << rhs_ << '\n';
  }


};

static Statement *parseSwapStatement(string const &token,
                                     string::const_iterator &first,
                                     string::const_iterator &last
                                    )
{
  if(token.size() != 3)
  {
    cerr << programName << " Error:  x-statement expects 2 digits, got:" << endl;
    cerr <<                "         '" << token << "'\n";
    exit(1);
  }

  int l = token[1] - '0';
  int r = token[2] - '0';

  return new swapStatement(l, r);


}
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct splitStatement
: 
public Statement
  //
  //.  a splitStatement splits a variable into 2 pieces:  the left hand side is kept
  //.  in the original variable, and the right hand side is moved into another variable.
  //.  There are two modes of operation:
  //.
  //.    splitting based on an explicit character count.
  //.    splitting based on a delimiter character.
  //.
  //.  When splitting on delimiters, a string is supplied.  All characters in the
  //.  string are assumed to be delimiters.  All characters up to the first delimiter
  //.  in the left hand side variable name are left alone.  The delimiter and everything
  //.  after it are removed.  The removed text, except for the initial delimiter are
  //.  oved into the right hand side variable.
  //
{
  int lhs_, mid_, rhs_;
  string delimiters_;
  size_t count_;

  splitStatement(string const &d, int l, int m, int r)
  : delimiters_(d),
    lhs_(l),
    mid_(m),
    rhs_(r),
    count_(0)
  {
  }

  splitStatement(int l, int r, int c)
  : delimiters_(""),
    lhs_(l),
    rhs_(r),
    count_( size_t(c) )
  {
  }

  int execute_v(istream &s, WhileContext* wc) const
  {
    string lhs = stringRegisters[lhs_];
    
    stringRegisters[rhs_].resize(0);

    if( delimiters_.size() == 0 )
    {
      if( count_ < lhs.size() )
      {
        // there is work to do

        stringRegisters[rhs_] = lhs.substr(count_, lhs.size() - count_);

        stringRegisters[lhs_].resize(count_);

      }

    }
    else
    {
      size_t pos = lhs.find_first_of(delimiters_);

      if(pos < lhs.size())
      {
        // there is something to do

        stringRegisters[lhs_] = lhs.substr(0, pos);

        stringRegisters[mid_] += lhs[pos];

        stringRegisters[rhs_] = lhs.substr(pos+1, lhs.size() - pos - 1);

      }
    }


    return ok;
  }

  void output(ostream &s) const
  {
    s << "+++S";

    if(delimiters_.size())
      s << "W";
    else
      s << "C";

    s << lhs_ << rhs_;

    if(delimiters_.size())
      s << " " << delimiters_;
    else
      s << " " << count_;
  }

};

//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct splitRegexStatement
: public Statement
{
  int from_;
  int delim_;
  int after_;

  SimpleRegex regex_;

  splitRegexStatement(int from, int delim, int after, string const &regex)
  : from_(from),
    delim_(delim),
    after_(after),
    regex_(regex)
  {
    if(!regex_.ok())
    {
      cerr << programName << "  Error:  SR-statement cannot compile the regex\n"
                          << "           expression='" << regex << "'\n";

      exit(1);
    }
  }

  int execute_v(istream &s, WhileContext* wc) const
  {
    string text = stringRegisters[from_];

    int matches = regex_(text);

    if(matches)
    {
      size_t offset = size_t (regex_.matches()[0].offset);
      size_t length = size_t (regex_.matches()[0].length);

      stringRegisters[delim_] = text.substr(offset, length);

      stringRegisters[after_] = text.substr(offset + length, text.size() - (offset + length) );

      stringRegisters[from_].resize(offset);
    }
    else
    {
      stringRegisters[after_].resize(0);
      stringRegisters[delim_].resize(0);
    }


    return ok;
  }

  void output(ostream &s) const
  {
    s << "+++SR" << from_ << delim_ << after_ << " /" << string(regex_) << "/\n";
  }



};

static Statement *parseSplitRegexStatement(string const &token, 
                                           string::const_iterator &first,
                                           string::const_iterator &last
                                          );


static Statement *parseSplitStatement(string const &token, 
                                      string::const_iterator &first,
                                      string::const_iterator &last
                                     )
{
  // A split statement looks like this:
  //
  //   SW012 :=\;*
  //   SC92 13
  //   SR012 /regex/
  //
  // See the splitStatement's comments, above

  bool parseError = false;

  if(token.size() != 5)
  {
    char secondChar = 0;

    if(token.size() > 2)
      secondChar = token[1];

    if(secondChar != 'C')
      parseError = true;

  }

  char mode = token.size() ? token[1] : 0;

  if(mode == 'R')
    return parseSplitRegexStatement(token, first, last);



  if( (mode != 'W') && (mode != 'C') )
  {
    parseError = true;
  }

  string secondToken;

  if(!parseError)
  {
    skipblanks(first, last);

    while(first != last)
    {
      if(*first == ';')
        break;
      else
      if(*first == '\\')
      {
        ++first;

        handleEscapeSequence(first, last, secondToken);

      }
      else
        secondToken += *first++;
    }
  }

  if(parseError || (secondToken.size() == 0) )
  {
      cerr << programName << " Error:  S-statement error, expected:\n"
           << "  SW012 =:\n"
           << "  SC85 31\n"
           << "but received '" << token << "'\n";
      logScriptPosition(cerr, first);
      exit(1);
  }

  int lhs = 0, mid=0, rhs = 0;

  if(mode == 'C')
  {
    lhs = token[2] - '0';
    rhs = token[3] - '0';
  }
  else
  {
    lhs = token[2] - '0';
    mid = token[3] - '0';
    rhs = token[4] - '0';
  }

  if(   (lhs < 0)
     || (lhs > 9)
     || (rhs < 0)
     || (rhs > 9)
     || (mid < 0)
     || (mid > 9)
    )
  {
        cerr << programName << " Error:  S-statement error, expected to digits for variable numbers\n"
             << "  found '" << token.substr(2,2) << "'\n";
        logScriptPosition(cerr, first);
        exit(1);
  }



  if(token[1] == 'C')
  {
    // validate that the character count is a number

    string::const_iterator a = secondToken.begin(), b = secondToken.end();

    while(a != b)
    {
      if(! isdigit(*a++) )
      {
        cerr << programName << " Error:  SC-statement error, expected a valid character count\n"
             << "  found '" << secondToken << "'\n";
        logScriptPosition(cerr, first);
        exit(1);
      }
    }

    return new splitStatement(lhs, rhs, atoi( secondToken.c_str() ) );

  }

  // ok, here we are dealing with a delimiter string

  return new splitStatement(secondToken, lhs, mid, rhs);

}

static Statement *parseSplitRegexStatement(string const &token, 
                                           string::const_iterator &first,
                                           string::const_iterator &last
                                          )
{
  if(    (token.size() != 5)
     || !( isdigit(token[2]) && isdigit(token[3]) && isdigit(token[4]))
    )
  {
    cerr << programName << " Error:  SR-statement error, expected SR[digit][digit][digit] /exp/\n";
    logScriptPosition(cerr, first);
    exit(1);
  }

  int from  = token[2] - '0';
  int delim = token[3] - '0';
  int after = token[4] - '0';

  skipblanks(first, last);

  string regex;

  int slashCount=0;


  while(first != last)
  {
    if(*first == '\\')
    {
      ++first;
      handleEscapeSequence(first, last, regex, true);
      continue;
    }


    if(*first == '/')
    {
      ++slashCount;

      if(!regex.empty())
      {
	++first;
        break;
      }

      ++first;

      continue;
        
    }

    if(!slashCount)
    {
      cerr << programName << " Error:  SR-statement error, expected SR[digit][digit][digit] /exp/\n";
      logScriptPosition(cerr, first);
      exit(1);
    }

    if(*first == '\n')
    {
      ++first;

      handleEscapeSequence(first, last, regex);
    }
    else
      regex += *first++;
  }

  if(!slashCount)
  {
      cerr << programName << " Error:  SR-statement error, expected SR[digit][digit][digit] /exp/\n";
      logScriptPosition(cerr, first);
      exit(1);
  }

  return new splitRegexStatement(from, delim, after, regex);




     
}





//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct cutStatement
: public Statement
{
  CutSet cutter_;
  int    variable_;
  string range_;

  cutStatement(string const &s, int variable)
  : cutter_(s),
    variable_(variable),
    range_(s)
  {
  }

  int execute_v(istream &s, WhileContext* wc) const
  {
    string text = stringRegisters[variable_];

    string resultant = cutter_(text);

    stringRegisters[variable_] = resultant;

    return ok;
  }

  void output(ostream &s) const
  {
    s << "+++c" << variable_ << " " << range_ << '\n';
  }

};


static Statement *parseCutStatement(string const &token,
                                    string::const_iterator &first,
                                    string::const_iterator &last
                                   )
{
  if(   (token.size() != 1) 
     && (token.size() != 2) 
     && ( (token.size() > 1) 
     && !isdigit(token[1]) ) 
    )
  {
    cerr << programName << " Error:  c-statement must look like one of these:\n"
                        << "           c range\n"
                        << "           c[digit] range\n"
                        << "         got this:\n"
                        << "           " << token
                        << endl;
    logScriptPosition(cerr, first);
    exit(1);
  }

  skipblanks(first, last);

  string range;
  int variable=0;

  if(token.size() >1)
    variable = token[1] - '0';

  while(first != last)
  {
    if( (*first == ';') || isspace(*first) )
      break;

    range += *first++;
  }

  // now validate characters in the range

  string::const_iterator a = range.begin(), b = range.end();


  while(a != b)
  {
     char c = *a++;

     if( !(
            isdigit(c) || (c == ',' ) || (c == '-' )
        
          )
       )
     {
        cerr << programName << " Error:  c-statement bad character in range:  '" << c << "'\n";
        logScriptPosition(cerr, first);
        exit(1);
     }

  }

  if(   (range.size() < 1)
     || !isdigit(range[range.size()-1])
     || !isdigit(range[0])
    )
  {
    cerr << programName << " Error:  c-statement bad character in range:  '" << range << "'\n"
                        << "         first and last characters must be digits\n";
    logScriptPosition(cerr, first);
    exit(1);
  }

  return new cutStatement(range, variable);


}

static void handleEscapeSequence(string::const_iterator &first,
                                 string::const_iterator &last,
                                 string                 &substString,
                                 bool                    leaveSlash
                                )
{
  //
  //.  This function handles the interpretation of the characters after
  //.  the '\' in an escape character sequence.  It does so by appending 
  //.  the appropriate text to substStrin.
  //.
  //.  Note that first is assumed to be positioned immediately after the
  //.  \ character and that the \ has not been appended to substString --
  //.  that is why you are calling this function, after all.
  //
  //   Maintenance note:  this function is called by the subst statement
  //   parser.  So, if you decide to support octal characters, you'll mess
  //   up the \0-\9 interpretation.  Supporting hex will be ok, but to do
  //   octal, would be bad.
  //

  if(first == last)
  {
    substString += '\\';   // \ at end of string
  }
  else
  if(*first == 'e')
  {
    // replace \e with 0x1b

    substString += char(0x1b);
    ++first;
  }
  else
  if(*first == 'n')
  {
    // replace \n with newline

    substString += '\n';
    ++first;
  }
  else
  if(*first == 't')
  {
    // replace \t with tab

    substString += char('\t');
    ++first;
  }
  else
  if(*first == 'b')
  {
    // replace \b with backspace

    substString += char(0x08);
    ++first;
  }
  else
  if(*first == 'r')
  {
    // replace \r with '\r'

    substString += '\r';
    ++first;
  }
  else
  if(*first == 'a')
  {
    // replace \a with bel

    substString += char(0x07);
    ++first;
  }
  else
  if(*first == 's')
  {
    // replace \s with ' '

    substString += ' ';
    ++first;
  }
  else
  {
    // let someone else handle the escape sequence
    if(leaveSlash)
      substString += '\\';

    substString += *first++;
  }
}

//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct fileStatement
: public Statement
{
  int variable_;

  Statement *child_;


  fileStatement(int var, Statement *block)
  : variable_(var),
    child_(block)
  {
  }

  int execute_v(istream &s, WhileContext* wc) const
  {
    string text = stringRegisters[variable_];

    ifstream file(text.c_str());

    if(!file.is_open())
    {
      cerr << programName << " Error:  F-statement error, cannot read file '" << text << "'" << endl;
      exit(1);
    }

    WhileContext fwc;

    while(file) // not end of input stream
    {
      int rc = child_->execute(file, &fwc);  // the child will be a blockStatement

      if(rc == returning)
        return returning;

      if(rc == breaking)
        break;

      // if the return code is contining or ok, just restart the block

    }

    return ok;
  }

  void output(ostream &s) const
  {
    s << "+++F" << variable_ << '\n' << (*child_);
  }

};

static Statement *parseFileStatement(string const &token,
                                     string::const_iterator &first,
                                     string::const_iterator &last
                                    )
{
  int variable= 0;

  if(!( 
          (token.size() == 1)
       || (
              (token.size() == 2)
           && isdigit(token[1])
          )
      )
    )
  {
    cerr << programName << " Error:  F-statement needs either 'F' or 'F[digit]', got '" << token << "'\n";
    logScriptPosition(cerr, first);
    exit(1);
  }

  if(token.size() == 2)
  {
    variable = token[1] - '0';
  }

  Statement *block = parseStatement(first, last);

  if(!block)
  {
    cerr << programName << " Error:  F-statement must include a script\n";
    logScriptPosition(cerr, first);
    exit(1);
  }

  return new fileStatement(variable, block);
}

//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct writeStatement
: public Statement
{
  int from_;
  int to_;


  writeStatement(int from, int to)
  : from_(from),
    to_(to)
  {
  }

  int execute_v(istream &s, WhileContext* wc) const
  {
    string text = stringRegisters[from_];
    string name = stringRegisters[to_];

    ofstream output(name.c_str());

    if(!output.is_open())
    {
      cerr << programName << " Error:  W-Statement cannot write to file '" << name << "'" << endl;
      exit(1);
    }

    output << text;

    return ok;
  }


  void output(ostream &s) const
  {
    s << "+++W" << from_ << to_ << '\n';
  }



};

static Statement *parseWriteStatement(string const           &token,
                                      string::const_iterator &first,
                                      string::const_iterator &last
                                     )
{

  if(   (token.size() != 3)
     || !( isdigit(token[1]) && isdigit(token[2]) )
    )
  {
    cerr << programName << " Error:  W-Statement must look like this:  W[digit][digit]\n";
    logScriptPosition(cerr, first);
    exit(1);
  }

  return new writeStatement(token[1]-'0', token[2]-'0');

}

//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct regPatternStatement
: public Statement
{
  //
  //  a regPatternStatement is a pattern conditional statement where
  //  the pattern is assumed to exist in a register rather than being
  //  a compile time constant.
  //

  int        regvar_;
  int        textvar_;
  bool       not_;
  Statement *child_;

  regPatternStatement(int regvar, int textvar, bool notv, Statement *child)
  : regvar_(regvar),
    textvar_(textvar),
    not_(notv),
    child_(child)
  {
  }

  int execute_v(istream &s, WhileContext* wc) const
  {
    string regex   = stringRegisters[regvar_];
    string operand = stringRegisters[textvar_];

    SimpleRegex reg(regex);

    if(!reg.ok())
    {
      cerr << programName << " Error:  cannot compile regular expression\n"
                          << " from register " << regvar_ << "\n"
                          << " text is       " << regex   << "\n"
                          ;
      exit(1);

    }

    int matches = reg(operand);

    if(not_)
    {
      if(matches)
        matches = 0;
      else
        matches=1;
    }

    if(matches)
    {
      return child_->execute_v(s,wc);
    }


    return ok;
  }



  void output(ostream &s) const
  {
    s << "+++|" << regvar_ << textvar_ ;

    if(not_)
      s << "!";

    s << '\n' << *child_;

  }

};

Statement *parseRegPatternStatement(string const &token,
                                    string::const_iterator &first,
                                    string::const_iterator &last
                                   )
{
  //
  //  The token should consist only of the the single letter |.
  //

  skipblanks(first,last);

  if(   (first == last)
     || !isdigit(*first)
    )
  {
    cerr << programName << " Error:  |-statement needs two digits and a command!\n";
    logScriptPosition(cerr, first);
    exit(1);
  }

  int regvar = *first++ - '0';

  if(   (first == last)
     || !isdigit(*first)
    )
  {
    cerr << programName << " Error:  |-statement needs two digits and a command!\n";
    logScriptPosition(cerr, first);
    exit(1);
  }

  int textvar = *first++ - '0';

  skipblanks(first, last);

  if(   (first == last) )
  {
    cerr << programName << " Error:  |-statement needs two digits and a command!\n";
    logScriptPosition(cerr, first);
    exit(1);
  }

  bool notvar = false;

  if(*first == '!')
  {
    notvar = true;

    ++first;

  }

  return new regPatternStatement(regvar, textvar, notvar, parseStatement(first,last));


}
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

static map<string, Statement*> functions;

struct functionStatement
: public Statement
{
  string     name_;
  Statement *body_;
  int        parms_; 

  functionStatement(string const &name, 
                    int           parms, // must be 0-9
                    Statement    *body
                   )
  : name_(name),
    body_(body),
    parms_(0)
  {
    map<string,Statement*>::iterator previous = functions.find(name);

    if(previous != functions.end())
    {
      cerr << programName << " Error:  duplicate definition of function " << name << endl;
      exit(1);
    }

    functions.insert( pair<string,Statement*>(name, this) );

  }


  int execute_v(istream &s, WhileContext* wc) const
  {
    // functions don't do anything -- call statements do!
    return ok;
  }

  void output(ostream &s) const
  {
    s << "+++Def " << name_ << parms_ << '\n' << (*body_);
  }


};

static Statement *parseFunctionStatement(string const &token,
                                         string::const_iterator &first,
                                         string::const_iterator &last
                                        )
{
  string name;
  string parms;

  gobbleToken(name, first, last);
  gobbleToken(parms, first, last);

  if(name.empty() || parms.empty() || !isalpha(name[0]) || !isdigit(parms[0]))
  {
    cerr << programName << " Error:  Def-statement expects a function name and parameter count, got:\n"
                        << "  name=" << name << '\n'
                        << "   cnt=" << parms 
                        << endl;
    logScriptPosition(cerr, first);
    exit(1);

  }

  int parmCount = atoi(parms.c_str());
  
  return new functionStatement(name, parmCount, parseStatement(first,last));

}
//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct callStatement
: public Statement
{
  string name_;
  char   parms_[10]; // an array of 10 booleans each mapping to a register -- specifying whether
                     // that register is included in the call.

  callStatement(string const &name, char const parms[10])
  : name_(name)
  {
    copy(parms, parms+10, parms_);
  }

  int execute_v(istream &s, WhileContext* wc) const
  {
    return handleCall(s);
  }

  void output(ostream &s) const
  {
    s << "+++C " << name_ << ' ';

    char const *first = parms_;
    char const *last  = first+10;

    char buffer[100];

    char *scan = buffer;

    while(first != last)
    {
      if(*first)
      {
        if(scan != buffer)
          *scan++ = ',';
        
        *scan++ = '0' + ( first - parms_ );
      }

      ++first;
    }

    *scan++ = 0;

    s << buffer << endl;

  }


  int handleCall(istream &s) const
  {
    //
    //  this function handles the function call -- it save/restores the global
    //  variable set, stringRegisters
    //

    map<string,Statement*>::iterator pos = functions.find(name_);

    if(pos == functions.end())
    {
      cerr << programName << " Error:  call to undefined function '" << name_ << "'\n";
      exit(1);
    }

    //
    // save the current register set
    //

    string savedRegisters[10];

    copy(stringRegisters, stringRegisters+10, savedRegisters);

    // first erase the global register set

    int i;

    for(i=0; i < 10; ++i)
      stringRegisters[i].clear();

    // then populate the registers, starting with 0 with the data
    // that is part of the call (1,3,5...)

    string* outreg = stringRegisters;

    for(i=0; i < 10; ++i)
    {
      if(parms_[i])
      {
        *outreg++ = savedRegisters[i];
      }
    }

    // call the function

    WhileContext fwc;

    functionStatement *funcStatement = dynamic_cast<functionStatement*>(pos->second);

    int rv = funcStatement->body_->execute(s, &fwc);

    // set up the returned registers

    outreg = stringRegisters;

    for(i=0; i < 10; ++i)
    {
      if(parms_[i])
      {
        savedRegisters[i] = *outreg++;
      }
    }

    copy(savedRegisters, savedRegisters+10, stringRegisters);


    // and return...

    if(rv == returning)
      rv = ok;

    return rv;

  }


};

static Statement *parseCallStatement(string const &token,
                                     string::const_iterator &first,
                                     string::const_iterator &last
                                    )
{
  string name;
  string parms;

  gobbleToken(name, first, last);

  skipblanks(first, last);

  while(first != last)
  {
    char c = *first;

    if(c == ',' || isdigit(c))
     parms += *first++;
    else
     break;
  }



  if(name.empty() || !isalpha(name[0]))
  {
    cerr << programName << " Error:  C-statement expects function name and register list\n";
    logScriptPosition(cerr, first);
    exit(1);
  }


  char parmSet[10];

  fill_n(parmSet, 10, char(0));

  char *scan = parmSet;

  string::const_iterator a = parms.begin(), b = parms.end();

  while(a != b)
  {
    char c = *a++;

    if(!isdigit(c) && (c != ','))
    {
      cerr << programName << " Error:  C-Statement register list badly formed:  " 
           << parms << '\n';

      logScriptPosition(cerr, first);
      exit(1);
    }

    if(c != ',')
    {
      int regnum = c - '0';

      if(parmSet[regnum])
      {
        cerr << programName << " Error:  C-Statement register list badly formed: register "
             << regnum << " is specified twice!" << endl;
  
        logScriptPosition(cerr, first);
        exit(1);
      }
      parmSet[c-'0'] = 1;
    }

  }

  return new callStatement(name, parmSet);

}



struct IfStatement
: public Statement
{
  SimpleRegex regex_;
  int         variable_;
  bool        not_;
  Statement*  ifClause_;
  Statement*  elseClause_;

  IfStatement(string const &expression,
              int           variable,
              bool          invertExpression,
              Statement    *ifClause,
              Statement    *elseClause
             )
  : regex_(expression),
    variable_(variable),
    not_(invertExpression),
    ifClause_(ifClause),
    elseClause_(elseClause)
  {
  }

  int execute_v(istream &s, WhileContext* wc) const
  {
    string text = stringRegisters[variable_];

    int matches = regex_(text);

    if(not_)
    {
      matches = ! matches;
    }

    if(matches && ifClause_)
      return ifClause_->execute_v(s, wc);

    if( (!matches) && elseClause_)
      return elseClause_->execute_v(s, wc);

    return ok;
  }

  void output(ostream &s) const
  {
    s << "+++if /" << (string const &)(regex_) << "/" << (not_ ? "!" : "" ) << endl;

    if(ifClause_)
      s << *ifClause_;

    s << "else\n";

    if(elseClause_)
      s << *elseClause_;

  }

};

static Statement *parseIfStatement(string const &token,
                                   string::const_iterator &first,
                                   string::const_iterator &last)
{
  // parse this syntax:
  //
  //   if /regex/~1! {cmd;} else {cmd; }

  skipblanks(first, last);

  if(first == last)
  {
    cerr << programName << " Error:  if-statement error:  expected /regex/" << endl;
    logScriptPosition(cerr,first);
    exit(1);
  }


  ++first; // first is now 1 past slash


  string::const_iterator slash = findInRange('/', first, last);


  if(slash == last)
  {
    cerr << programName << " Error:  expected trailing / in pattern conditional statement" << endl;
    logScriptPosition(cerr,first); 
    exit(1);
  }

  string expression;

  expression.append(first, slash);

  first = slash;

  ++first; // first now past the second slash

  if(first == last)
  {
    cerr << programName << " Error:  if-statement expected statement after /expr/" << endl;
    logScriptPosition(cerr,first); exit(1);
    
  }

  int variable = 0;

  if(*first == '~')
  {
    ++first;

    if(first == last)
    {
      cerr << programName << " Error:  if-statement expected digit after /expr/~" << endl;
      logScriptPosition(cerr,first); 
      exit(1);
    }

    if(!isdigit(*first))
    {
      cerr << programName << " Error:  if-statement expected variable number after /expr/~" << endl;
      logScriptPosition(cerr,first); 
      exit(1);
    }

    variable = *first - '0';

    ++first;
    
  }
  

  bool invertedLogic = false;

  if(*first == '!')
  {
    invertedLogic = true;
    ++first;

  }

  skipblanks(first, last);

  if(first == last)
  {
    cerr << programName << " Error:  if-statement expected statement after /expr/" << endl;
    logScriptPosition(cerr,first); 
    exit(1);
  }

  Statement *ifClause = parseStatement(first, last);

  if(!ifClause)
  {
    cerr << programName << " Error:  if-statement expected statement after /expr/" << endl;
    logScriptPosition(cerr,first); 
    exit(1);
    
  }

  while(first != last)
  {
    if(isspace(*first) || (*first == ';') )
      ++first;
    else
      break;

  }

  Statement *elseClause=0;

  if( (first != last) && (*first == 'e') )
  {
    string elseToken;

    string::const_iterator savedFirst = first;

    gobbleToken(elseToken, first, last);

    static string elsekeyword("else");

    if(elseToken != elsekeyword)
    {
      first = savedFirst;
      --first;
    }
    else
    {
      elseClause = parseStatement(first, last);
    }

  }

  return new IfStatement(expression, variable, invertedLogic, ifClause, elseClause);


}


//////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

struct ReadAllStatement
: public Statement
{
  int    variable_;
  int    fileNameVar_;
  string fileNameText_;


  ReadAllStatement(int var, int fv, string fn)
  : variable_(var),
    fileNameVar_(fv),
    fileNameText_(fn)
  {
  }

  int execute_v(istream &s, WhileContext *wc) const
  {
    string filename;

    if(fileNameVar_ >= 0)
    {
      filename = stringRegisters[fileNameVar_];
    }
    else
    if(!fileNameText_.empty())
    {
      filename = fileNameText_;
    }

    if(!filename.empty())
    {
      // the user has specified a filename

      ifstream file(filename.c_str());

      if(!file.is_open())
      {
        cerr << programName << " Error:  R statement read error in file '" << filename << "'" << endl;
        exit(1);
      }

      getline(file, stringRegisters[variable_], char(0));
    }
    else
    {
      getline(s, stringRegisters[variable_], char(0));
    }


    return ok;
  }

  void output(ostream &s) const
  {
    s << "+++R" << variable_;
    
    if(fileNameVar_ >= 0)
      s << fileNameVar_;

    s << " " << fileNameText_;

    s << '\n';

  }

};

static Statement *parseReadAllStatement(string const &token,
                                        string::const_iterator &first,
                                        string::const_iterator &last
                                       )
{
  int variable=0;
  int fileNameVar = -1;
  string filename;

  if(   token.size() > 3   
     || (token.size() > 1 && !isdigit(token[1]))
     ||
     (
       token.size() == 3 && !isdigit(token[2])
     )
    )
  {
    cerr << programName << " Error:  R-statement expects one of the following forms:\n"
                        << "         R;\n"
                        << "         R1 filename;\n"
                        << "         R23;\n"
                        << endl;
    logScriptPosition(cerr,first); 
    exit(1);
   
  }

  if(token.size() < 3)
  {
    // we are expecteding R4 filename;

    skipblanks(first, last);

    while(first != last)
    {
      // read the filename -- allowing most characters through -- but stopping on an
      // un-protected ;, }, |, etc.

      if(*first == '\\')
      {
        ++first;
      }
      else
      if(*first == ';' || *first == '}' || isspace(*first) )
        break;

      if(first != last)
        filename += *first++;
    }

    if(token.size() > 1)
      variable = token[1] - '0';
  }
  else
  if(token.size() == 3)
  {
    variable    = token[1] - '0';
    fileNameVar = token[2] - '0';
  }
  else
  {
      cerr << programName << " bug:  " << __FILE__ << ":" << __LINE__ << endl;
      logScriptPosition(cerr,first); 
      exit(1);
  }


  return new ReadAllStatement(variable, fileNameVar, filename);

}


static string logText(string const &s, size_t maxLength)
{
  // convert input string into "loggable text" which has no embedded control
  // characters and less than about maxLength characters long

  string rv;

  string::const_iterator first = s.begin(), last = s.end();

  while(   (rv.size() < maxLength)
        && (first != last)
       )
  {
    switch(*first)
    {
      case '\t': rv += "\\t"; break;
      case '\n': rv += "\\n"; break;
      case '\r': rv += "\\n"; break;
      case '\f': rv += "\\f"; break;
      case '\a': rv += "\\a"; break;
      default:
        rv += *first;
    }

    ++first;
  }

  return rv;
}


struct UnReadStatement
: public Statement
{
  int variable_;

  UnReadStatement(int var)
  : variable_(var)
  {
  }

  int execute_v(istream &s, WhileContext *wc) const
  {
    string text = stringRegisters[variable_];

    pushedBackLines.push_back(text);  // does not effect input line number!

    return ok;
  }

  void output(ostream &s) const
  {
    s << "+++N" << variable_ << '\n';
  }

};

static Statement *parseUnReadStatement(string const           &token, 
                                       string::const_iterator &first,
                                       string::const_iterator &last
                                      )
{
  int variable = 0;

  if(token.size() > 1)
  {
    int err=false;

    if(isdigit(token[1]))
      variable = token[1] - '0';
    else
      err = true;

    if(   (token.size() > 2) 
       || err
      )
    {
      cerr << programName << " Error:  invalid N token:  expected N or N[digit]"
           << endl;

      logScriptPosition(cerr,first);

      exit(1);
      
    }
    
  }

  return new UnReadStatement(variable);


}

