#ifndef REGEX_HEADER_INCLUDED
#define REGEX_HEADER_INCLUDED


//
// Copyright 2002, 2007 Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.




#include <string>
#include <regex.h>

//@File
//
//. This file defines a C++ wrapper around the standard regular
//. expression functionality.

class SimpleRegex
    //
    //. A simplistic interface to the regular expression.  This is meant
    //. to only work with short string -- 200 characters or less.
    //. Construction equals compilation as does copy construction and
    //. assignment.  To search a string, first construct a SimpleRegex
    //. then treat it like a function and call it.  See the operator() methods
    //. for variants.  The operators return the number of matches.  Errors
    //. and matched items are mutable -- so just because one of these things
    //. is declared const, doesn't mean the error message and the matched
    //. items remain inviolate.
{
public:

  struct match     //. match  a struct describing a sub-string that matched
		   //. the pattern.
  {
      int offset;  /* location in the 'string' where the regexp was found */
      int length;
  };

private:

  regex_t compiled_expression; //. The real @@::regex_t@@ object that does
			       //. the work.

  std::string expression_;     //. saved for later use -- in copying

  enum constants_
  {
    max_matches=20
  };

  mutable match matches_[max_matches]; //. buffer hold matches from a given search

  mutable char const *error_;          //. construction error message if any

  bool ok_;                            //. compilation was ok, not searching

  mutable int match_count_;	       //. set when an operator	function is called
				       //. set to 0 when constructed or copied!
				
  std::string expand(std::string const&, std::string const&) const;
    //. used in the substitute function as a helper.

public:


  SimpleRegex(std::string const &expression="");  //- SimpleRegex()  Construct from a std::string
  SimpleRegex(char const *expression);            //- SimpleRegex()  Construct from a char *
  SimpleRegex(SimpleRegex const &r);              //- SimpleRegex()  Copy construct

  SimpleRegex &operator=(SimpleRegex const &r); //- operator= copy from another SimpleRegex
  SimpleRegex &operator=(std::string const &r); //- operator= copy from std::string
  SimpleRegex &operator=(char const *r);        //- operator= copy from char *

  // the following operators return the number of matches.
  int operator() (char const *string, int stringLength) const;
  int operator() (std::string const &s,
		  int offset=0,
		  int length=0x7ffffff) const;
   //. Check a string or character buffer for a match
   //. and return the number of matches.  See matches() below.

  bool ok() const { return ok_ && error_ == 0; } //- ok()  regex was valid?

  char const *error() const { return error_; } //- error()  text of error message

  bool anchored() const;
    //.  returns true if the regular expression is anchored to the beginning of the string

  bool empty() const { return expression_.empty(); }
    //.  returns true if the expression is an empty string

  operator std::string const &() const { return expression_; } //- what was the regex anyway?

  ~SimpleRegex();  //- ~SimpleRegex() destructor

  match const *matches() const { return matches_; }  //- matches() array of matches found

  std::string substitute(std::string const &original,
			 std::string const &replacement) const;
    //.  This function replace the text in the original string which is
    //.  identified by the match array of this regular expression with a
    //.  value computed from the replacement string.  Here's how you use
    //.  the substitute method:
    //.
    //.    1.  construct a regular expression which can be used to perform
    //.        regex find.
    //.
    //.    2.  execute this regular expression on some string value.
    //.
    //.    3.  if a match was found, use the substitute method to compute a
    //.        a new string which is a copy of the comparand string value
    //.        with the matched parts replaced by a new string computed from
    //.        'replacement'.
    //.
    //.  The replacement string is used to compute a replacement value and
    //.  contain literal text and also can contain \0 ... \9.  When these
    //.  substrings occur, they are replaced with text from the original
    //.  comparand string.  See the man page for ex, sed, perl, etc for
    //.  regular expression substitution logic.

  bool replace(std::string       &rv, 
               std::string const &source, 
               std::string const &replacement, 
               int           maxIterations=0x7fffffff
              ) const;
    //.  Replace a pattern defined by this SimpleRegex within a source string
    //.  with a replacement at most maxReplacements times and return true if at least one replacement
    //.  occurred.

};



#endif
